"""
Created on Wed Jan 16 18:02:25 2019

@author: lpaone
"""
import math
from dataclasses import field
from enum import Enum
from typing import Any, List, Optional, Tuple, Union

import numpy as np
from pydantic import BaseModel

import pycivil.EXAStructural.lawcodes.codeEC212 as fireCode
import pycivil.EXAStructural.codes as codes
import pycivil.EXAStructural.materials as materials
from pycivil.EXAGeometry.geometry import Point2d
from pycivil.EXAGeometry.shapes import Shape, ShapeArea, ShapeRect
from pycivil.EXAStructural.materials import ConcreteModel, SteelModel


class ShapeEnum(str, Enum):
    RECT = "rectangular"
    T_SHAPE = "T-shape"
    I_SHAPE = "I-shape"
    POLYGONAL = "polygon"


class ShapeModel(BaseModel):
    tp: Optional[ShapeEnum] = None
    id: int = 0
    descr: str = ""
    selected: List[int] = field(default_factory=list)


class RectangularShape(ShapeModel):
    tp: ShapeEnum = ShapeEnum.RECT
    width: float
    height: float

class TShape(ShapeModel):
    tp: ShapeEnum = ShapeEnum.T_SHAPE
    width: float
    height: float
    slab_width: float
    slab_height: float
    web_position: float

class RebarDisposerEnum(str, Enum):
    SINGLE = "single"
    LINE = "line"
    STIRRUP = "stirrup"
    STIRRUP_Y = "stirrup_single_leg_y"


class ShapePositionEnum(str, Enum):
    MT = "MT"
    MB = "MB"


class SteelDisposerSingle(BaseModel):
    id: int = -1
    tp: RebarDisposerEnum = RebarDisposerEnum.SINGLE
    area: float
    xpos: float
    ypos: float


class SteelDisposerOnLine(BaseModel):
    tp: RebarDisposerEnum = RebarDisposerEnum.LINE
    fromPos: ShapePositionEnum = ShapePositionEnum.MB
    diameter: float = 20
    steelInterDistance: float = 20
    distanceFromPos: float = 20
    number: float | int = 4


class SteelDisposerStirrup(BaseModel):
    tp: RebarDisposerEnum = RebarDisposerEnum.STIRRUP
    area: Optional[float] = None
    step: Optional[float] = None
    angle: Optional[float] = None


class SteelDisposerStirrupSingleLeg(BaseModel):
    id: int = -1
    tp: RebarDisposerEnum = RebarDisposerEnum.STIRRUP_Y
    diameter: float = 8
    angle: float = 0
    slope: float = 90
    centerX: float = 0
    centerY: float = 0
    length: float = 0
    step: float = 0


class AreaProperties(BaseModel):
    area: float = 0
    sx: float = 0
    sy: float = 0
    ixx: float = 0
    iyy: float = 0
    ixy: float = 0
    ixxx: float = 0
    ixxy: float = 0


class SectionStates(str, Enum):
    UNKNOWN = "UNKNOWN"
    COMPRESSED = "COMPRESSED"
    STRETCHED = "STRETCHED"
    PARTIALIZED = "PARTIALIZED"


class StructSectionItem:
    def __init__(
        self,
        shape: Union[Shape, None],
        material: Union[materials.Material, None] = None,
        ids: int = -1,
    ):
        if shape is None:
            shape = Shape()
        if material is None:
            material = materials.Material()

        self.__id = ids
        self.shape = shape
        self.material = material

    def setId(self, ids: Union[int, None]) -> None:
        self.__id = ids

    def getId(self) -> Union[int, None]:
        return self.__id

    def getShape(self):
        return self.shape

    def getOrigin(self):
        return self.shape.getOrigin()

    def setOriginX(self, x):
        p = self.shape.getOrigin()
        p.x = x
        self.shape.setOrigin(p)

    def setOriginY(self, y):
        p = self.shape.getOrigin()
        p.y = y
        self.shape.setOrigin(p)

    def getArea(self) -> float:
        return self.shape.getArea()

    def getDiameter(self) -> float:
        return self.shape.getDiameter()

    def getMaterial(self):
        return self.material

    def __str__(self):
        dispstr = "StructSectionItem Object: \n"
        dispstr = dispstr + "--------------------  \n"
        dispstr = dispstr + "Shape embedded-->\n"
        dispstr = dispstr + self.shape.__str__().replace("\n", "\n  | ")
        dispstr = dispstr + "Material embedded-->\n"
        dispstr = dispstr + self.material.__str__().replace("\n", "\n  | ")
        return dispstr


class ConcreteSection:
    """Class that instantiate a general concrete section

    Reference for geometry is:

                     ----------------
                     |              |
                     |              |
                     |    Y ^       |
                     |      |       |
                     |      |   X   |
                     |      O--->   |
                     |              |
                     |              |
                     |              |
                     ----------------

     for forces and moments (N compression negative)

                     ----------------
                     |              |
                     |              |
                     |      --.     |
                     |  My     \\    |
                     |  <<--O   ;   |
                     |      |  /    |
                     |     <--'     |
                     |  N = Fx < 0  |
                     |              |
                     ----------------


    """

    def __init__(self, ids: int = 0, descr: str = ""):

        self.__ids: int = ids
        self.__descr: str = descr

        # StructuralItemContainer
        self.__concreteItem: Union[StructSectionItem, None] = None
        self.__steelItems: list[StructSectionItem] = []
        self.__steelItemsWeights = []  # Weight to apply alpha for design
        self.__steelItemsAlphaWeight = 1.0  # Alpha designer factor
        self.__steelItemsTemperatures: List[float] = []  # Temperature for steel
        self.__steelItems_kappa_s: List[float] = []  # Temperature for steel

        # Coding by
        self.__code = codes.Code("EC2")
        self.__elementType = "beam"

        # Concrete material set
        cls_material = materials.Concrete(descr="EC2-C20/25 [default]")
        cls_material.setByCode(self.__code, "C20/25")
        self.__concreteMaterial = cls_material
        # Steel material set
        steel_material = materials.ConcreteSteel(descr="EC2-B450C [default]")
        steel_material.setByCode(self.__code, "B450C")
        self.__steelMaterial = steel_material

        self.__homogenization = 15.0

        # Material only for section
        self.__steelMaxStress = 0.85 * 450
        self.__steel_safetyFactorForULS = 1.15
        self.__concreteMaxStress = 0.65 * 20
        self.__cls_safetyFactorForULS = 1.50

        self.__useSectionMaterials = True

        # SLS stress and forces
        self.__steelStress: List[float] = []
        self.__concrStress: List[float] = []
        self.__concrDeform: List[float] = []
        self.__N = None
        self.__Mx = None
        self.__My = None

        # Steel for stippupts
        self.__steelStirruptArea = None
        self.__steelStirruptDist = None
        self.__steelStirruptAngle = None

    def setForces(self, N=None, Mx=None, My=None, M=None):
        if N is not None:
            self.__N = N
        if Mx is not None:
            self.__Mx = Mx
        if My is not None:
            self.__My = My
        if M is not None:
            self.__Mx = M

    def getForceN(self):
        return self.__N

    def getForceM(self):
        return self.__Mx

    def setCode(self, c):
        if isinstance(c, codes.Code):
            self.__code = c
        else:
            raise Exception("Argument must have [Code] object !!!")

    def setElementType(self, elTypeStr):
        if isinstance(elTypeStr, str):
            self.__elementType = elTypeStr
        else:
            raise Exception("Argument must be str type !!!")

    def setStructConcrItem(self, strusecitem: StructSectionItem):
        if isinstance(strusecitem, StructSectionItem):
            if isinstance(strusecitem.getShape(), Shape) and isinstance(
                strusecitem.getMaterial(), materials.Concrete
            ):
                self.__concreteItem = strusecitem
            else:
                raise Exception(
                    "Argument must have [Shape] and [Concrete] as material assigned !!!"
                )
        else:
            raise Exception("Argument must be [structSectionItem] !!!")

    def setStructSteelItems(self, strusecitems: List[StructSectionItem]):
        if isinstance(strusecitems, list):
            for item in strusecitems:
                if not isinstance(item, StructSectionItem):
                    raise Exception("Only list of [structSectionItem] !!!")
                if not (
                    isinstance(item.getShape(), Shape)
                    and isinstance(item.getMaterial(), materials.ConcreteSteel)
                ):
                    raise Exception(
                        "List item must have [Shape] and [Steel] as material assigned !!!"
                    )
            self.__steelItems = strusecitems
            self.__steelItemsWeights = len(strusecitems) * [1.0]
        else:
            raise Exception("Argument must be list of [structSectionItem] !!!")

    def setSteelWeights(self, weights):
        if not isinstance(weights, list):
            raise Exception("Only list for weights !!!")
        for i in weights:
            if not isinstance(i, float):
                raise Exception("Only float for weights list !!!")
        if not len(weights) == len(self.__steelItems):
            raise Exception(
                "Weights lenght must be %1.0i same of steel items !!!"
                % len(self.__steelItems)
            )
        self.__steelItemsWeights = weights  # finally

    def setSteelTemperatures(self, temperatures: List[float]):
        if len(temperatures) != len(self.__steelItems):
            raise Exception("Temperatures must have same lenght of steel items !!!")
        self.__steelItemsTemperatures = temperatures
        return True

    def setSteelAlphaWeight(self, weight):
        if isinstance(weight, float):
            self.__steelItemsAlphaWeight = weight
        else:
            raise Exception("Only float type for weight !!!")

    def getSteelFireFactors(self) -> List[float]:
        return self.__steelItems_kappa_s

    def getSteelTemperatures(self) -> List[float]:
        return self.__steelItemsTemperatures

    def getSteelStress(self) -> List[float]:
        return self.__steelStress

    def setSteelStress(self, s: List[float]):
        self.__steelStress = s

    def getConcrStress(self) -> List[float]:
        return self.__concrStress

    def setConcrStress(self, s: List[float]):
        self.__concrStress = s

    def getConcrDeform(self) -> List[float]:
        return self.__concrDeform

    def setConcrDeform(self, s: List[float]):
        self.__concrDeform = s

    def getDescr(self):
        return self.__descr

    def getId(self):
        return self.__ids

    def getElementType(self):
        return self.__elementType

    def getSteelRebar(self) -> list[StructSectionItem]:
        return self.__steelItems

    def getSteelAreaAt(self, i, alphaWeight=False, weights=False):
        if alphaWeight:
            used_alphaWeight = self.__steelItemsAlphaWeight
        else:
            used_alphaWeight = 1.0

        if weights:
            used_weights = self.__steelItemsWeights[i]
        else:
            used_weights = 1.0

        return self.__steelItems[i].getArea() * used_alphaWeight * used_weights

    def getSteelDiamAt(self, i) -> float:
        # return round(2 * math.sqrt(self.__steelItems[i].getArea() / math.pi))
        return 2 * math.sqrt(self.__steelItems[i].getArea() / math.pi)

    def mirrorXaxis(self):
        for s in self.__steelItems:
            origin = s.getShape().getOrigin()
            origin.y = -origin.y

    def calProp_As(self, alphaWeight=False, weights=False):
        return self.calSteelArea(alphaWeight, weights)

    def calProp_Ac(self):
        return self.calConcreteArea()

    def calProp_Ah(self, alphaWeight=False, weights=False):
        return (
            self.__homogenization * self.calSteelArea(alphaWeight, weights)
            + self.calProp_Ac()
        )

    def calProp_Ssx(self, alphaWeight=False, weights=False):
        if weights:
            using_weights = self.__steelItemsWeights
        else:
            using_weights = len(self.__steelItemsWeights) * [1]

        if alphaWeight:
            using_alphaWeight = self.__steelItemsAlphaWeight
        else:
            using_alphaWeight = 1.0

        Sx = 0.0
        for i, s in enumerate(self.__steelItems):
            Sx = (
                Sx
                + s.getShape().getArea()
                * using_weights[i]
                * using_alphaWeight
                * s.getShape().getShapePoint("G").y
            )

        return Sx

    def calProp_Ssy(self, alphaWeight=False, weights=False):
        if weights:
            using_weights = self.__steelItemsWeights
        else:
            using_weights = len(self.__steelItemsWeights) * [1]

        if alphaWeight:
            using_alphaWeight = self.__steelItemsAlphaWeight
        else:
            using_alphaWeight = 1.0

        Sy = 0.0
        for i, s in enumerate(self.__steelItems):
            Sy = (
                Sy
                + s.getShape().getArea()
                * using_weights[i]
                * using_alphaWeight
                * s.getShape().getShapePoint("G").x
            )

        return Sy

    def calProp_Scx(self):
        Scx = (
            self.__concreteItem.getShape().getShapePoint("O").y
            * self.__concreteItem.getShape().getArea()
        )
        return Scx

    def calProp_Scy(self):
        Scy = (
            self.__concreteItem.getShape().getShapePoint("O").x
            * self.__concreteItem.getShape().getArea()
        )
        return Scy

    def calProp_Shx(self, alphaWeight=False, weights=False):
        return (
            self.calProp_Scx()
            + self.calProp_Ssx(alphaWeight, weights) * self.__homogenization
        )

    def calProp_Shy(self, alphaWeight=False, weights=False):
        return (
            self.calProp_Scy()
            + self.calProp_Ssy(alphaWeight, weights) * self.__homogenization
        )

    def calBarycenterOfSteel(self, alphaWeight=False, weights=False):
        xg = self.calProp_Ssy(alphaWeight, weights) / self.calProp_As(
            alphaWeight, weights
        )
        yg = self.calProp_Ssx(alphaWeight, weights) / self.calProp_As(
            alphaWeight, weights
        )
        return Point2d(xg, yg)

    def calBarycenterOfConcrete(self) -> Point2d:
        return self.__concreteItem.getShape().getShapePoint("G")

    def calBarycenter(self, alphaWeight=False, weights=False):

        steelg = self.calBarycenterOfSteel(alphaWeight, weights)
        steelArea = self.calSteelArea(alphaWeight, weights)
        concrg = self.calBarycenterOfConcrete()
        concrArea = self.calConcreteArea()

        xg = (steelg.x * steelArea * self.__homogenization + concrg.x * concrArea) / (
            steelArea * self.__homogenization + concrArea
        )
        yg = (steelg.y * steelArea * self.__homogenization + concrg.y * concrArea) / (
            steelArea * self.__homogenization + concrArea
        )

        return Point2d(xg, yg)

    def calProp_Isx(self, alphaWeight=False, weights=False, yg: float = 0.0):
        if weights:
            using_weights = self.__steelItemsWeights
        else:
            using_weights = len(self.__steelItemsWeights) * [1]

        if alphaWeight:
            using_alphaWeight = self.__steelItemsAlphaWeight
        else:
            using_alphaWeight = 1.0

        Isx = 0.0
        for i, s in enumerate(self.__steelItems):
            area = s.getShape().getArea() * using_weights[i] * using_alphaWeight
            y = s.getShape().getShapePoint("G").y
            Isx = Isx + area * (y - yg) * (y - yg)
        return Isx

    def calProp_Icx(self, yg: float = 0.0):
        B = self.__concreteItem.getShape().w()  # type: ignore
        H = self.__concreteItem.getShape().h()  # type: ignore
        return 1 / 12 * B * H * H * H + B * H * yg**2

    def calProp_Ihx(
        self, alphaWeight: bool = False, weights: bool = False, barycenter: bool = False
    ) -> float:
        """
        Function calculates moment of inertia along x-axis where x-axis is horizontal axis that
        goes through xhg-axis. This axis is the barycenter axis of homogeneized section
        Args:
            barycenter: if True prop will be calculated on barycentric axis
            alphaWeight: moltiplication factor for all rebars
            weights: TODO

        Returns:
            float: barycenter axis of homogeneized section
        """
        if barycenter:
            yg = self.calBarycenter().y
        else:
            yg = 0.0

        return self.__homogenization * self.calProp_Isx(
            alphaWeight, weights, yg
        ) + self.calProp_Icx(yg)

    def calProp_Isy(self, alphaWeight=False, weights=False):
        if weights:
            using_weights = self.__steelItemsWeights
        else:
            using_weights = len(self.__steelItemsWeights) * [1]

        if alphaWeight:
            using_alphaWeight = self.__steelItemsAlphaWeight
        else:
            using_alphaWeight = 1.0

        Isy = 0.0
        for i, s in enumerate(self.__steelItems):
            area = s.getShape().getArea() * using_weights[i] * using_alphaWeight
            x = s.getShape().getShapePoint("G").x
            Isy = Isy + area * x * x
        return Isy

    def calProp_Icy(self):
        B = self.__concreteItem.getShape().w()  # type: ignore
        H = self.__concreteItem.getShape().h()  # type: ignore
        return 1 / 12 * H * B * B * B

    def calProp_Ihy(self, alphaWeight=False, weights=False):
        return (
            self.__homogenization * self.calProp_Isy(alphaWeight, weights)
            + self.calProp_Icy()
        )

    def getCode(self):
        return self.__code

    def calConcreteArea(self):
        return self.__concreteItem.getShape().getArea()

    def calSteelArea(self, alphaWeight=False, weights=False):
        if weights:
            using_weights = self.__steelItemsWeights
        else:
            using_weights = len(self.__steelItemsWeights) * [1]

        if alphaWeight:
            using_alphaWeight = self.__steelItemsAlphaWeight
        else:
            using_alphaWeight = 1.0

        area = 0.0
        for i, s in enumerate(self.__steelItems):
            area = area + s.getShape().getArea() * using_weights[i] * using_alphaWeight
        return area

    def calSteelAreaMin(self, Ned=0.0):
        if self.getCode().codeStr() == "EC2":
            if self.__elementType == "beam":
                fctm = self.__concreteMaterial.get_fctm()
                fyk = self.__steelMaterial.get_fsy()
                b = self.__concreteItem.getShape().w()  # type: ignore

                if len(self.__steelItems) != 0:
                    d = self.__concreteItem.getShape().h() - self.getSteelBotRecover()  # type: ignore
                else:
                    d = self.__concreteItem.getShape().h()  # type: ignore

                asmin = max([0.26 * fctm / fyk, 0.0013 * b * d])
            elif self.__elementType == "column":
                fyd = self.__steelMaterial.get_fsy() / self.__steel_safetyFactorForULS
                Ac = self.__concreteItem.getShape().getArea()
                asmin = max([0.10 * Ned / fyd, 0.002 * Ac])
            else:
                raise Exception("Element type unknown !!!")
        else:
            raise Exception("Actually only for EC2 code !!!")
        return asmin

    def calSteelAreaMax(self):
        if self.getCode().codeStr() == "EC2":
            asmax = 0.04 * self.__concreteItem.getShape().getArea()
        else:
            raise Exception("Actually only for EC2 code !!!")
        return asmax

    def addSteelArea(self, item, weight=1.0) -> Union[int, None]:
        if not isinstance(item, StructSectionItem):
            raise Exception("Only list of [structSectionItem] !!!")
        if not (
            isinstance(item.getShape(), Shape)
            and isinstance(item.getMaterial(), materials.ConcreteSteel)
        ):
            raise Exception(
                "List item must have [Shape] and [Steel] as material assigned !!!"
            )
        self.__steelItems.append(item)
        self.__steelItemsWeights.append(weight)
        return len(self.__steelItems) - 1

    def translateSteelItems(
        self, pStart: Union[Point2d, None], pEnd: Union[Point2d, None]
    ):
        if pStart is None:
            pStart = Point2d()
        if pEnd is None:
            pEnd = Point2d()

        for s in self.__steelItems:
            s.getShape().translate(pStart, pEnd)

    def translateConcreteItem(
        self, pStart: Union[Point2d, None], pEnd: Union[Point2d, None]
    ):
        if pStart is None:
            pStart = Point2d()
        if pEnd is None:
            pEnd = Point2d()
        self.__concreteItem.getShape().translate(pStart, pEnd)

    def getSteelWeights(self):
        return self.__steelItemsWeights

    def getSteelAlphaWeight(self):
        return self.__steelItemsAlphaWeight

    def getStructConcretelItem(self):
        return self.__concreteItem

    def getStructSteelItems(self):
        return self.__steelItems

    def findLowSteelItem(self):
        vals = []
        for item in self.__steelItems:
            vals.append(item.getShape().getOrigin().y)

        # Steel not present
        if len(vals) == 0:
            return StructSectionItem(ShapeArea(0.0))
        else:
            return self.__steelItems[vals.index(min(vals))]

    def findHitSteelItem(self):
        vals = []
        for item in self.__steelItems:
            vals.append(item.getShape().getOrigin().y)

        if len(vals) == 0:
            return StructSectionItem(ShapeArea(0.0))
        else:
            return self.__steelItems[vals.index(max(vals))]

    def getConcrVertexMaxInY(self):
        return self.__concreteItem.getShape().vertexMaxInY()[1]

    def getSteelTopRecover(self) -> float:
        top = self.__concreteItem.getShape().vertexMaxInY()[1]
        bot = self.findHitSteelItem().getShape().getOrigin()
        return (top - bot).y

    def getSteelBotRecover(self) -> float:
        bot = self.__concreteItem.getShape().vertexMinInY()[1]
        top = self.findLowSteelItem().getShape().getOrigin()
        return (top - bot).y

    def getSteelMaterial(self):
        return self.__steelMaterial

    def setConcreteMaterial(self, mat, concreteMaxStress=None):
        if isinstance(mat, materials.Concrete):
            self.__concreteMaterial = mat
        else:
            raise Exception("You need get in Concrete naterial object !!!")
        if concreteMaxStress is not None:
            if isinstance(concreteMaxStress, float):
                self.__concreteMaxStress = concreteMaxStress
            else:
                raise Exception("Concrete max stress must be float type !!!")

    def getConcreteMaterial(self):
        return self.__concreteMaterial

    def setSteelMaterial(self, mat, steelMaxStress=None):
        if isinstance(mat, materials.ConcreteSteel):
            self.__steelMaterial = mat
        else:
            raise Exception("You need get in ConcreteSteel naterial object !!!")
        if steelMaxStress is not None:
            if isinstance(steelMaxStress, float):
                self.__steelMaxStress = steelMaxStress
            else:
                raise Exception("Steel max stress must be float type !!!")

    def getSteelMaxStress(self) -> float:
        return self.__steelMaxStress

    def getSteelSafetyFactorForULS(self) -> float:
        return self.__steel_safetyFactorForULS

    def getConcreteMaxStress(self) -> float:
        return self.__concreteMaxStress

    def getClsSafetyFactorForULS(self) -> float:
        return self.__cls_safetyFactorForULS

    def getUseSectionMaterials(self) -> bool:
        return self.__useSectionMaterials

    def getHomogenization(self):
        return self.__homogenization

    def setHomogenization(self, n):
        if n is not None:
            if isinstance(n, float):
                self.__homogenization = n
            else:
                raise Exception("Homogeneization must be float type !!!")

    def getN(self) -> Union[float, None]:
        return self.__N

    def getMx(self) -> Union[float, None]:
        return self.__Mx

    def getMy(self) -> Union[float, None]:
        return self.__My

    def build2dInteractionCompleteDomain(
        self, nbPoints=100, SLS=False, negative_compression=False
    ) -> Any:

        if not SLS:
            NxMz_section, Fields, Bounding = self.build2dInteractionDomain(
                nbPoints, negative_compression
            )
        else:
            NxMz_section, Fields, Bounding = self.build2dInteractionDomainSLS(
                nbPoints, negative_compression
            )

        # TODO: this need to be fixed for barycenter not (0,0)
        for si in self.__steelItems:
            si.setOriginY(-si.getOrigin().y)

        if not SLS:
            (
                NxMz_section_flipped,
                Fields_flipped,
                Bounding_flipped,
            ) = self.build2dInteractionDomain(nbPoints, negative_compression)
        else:
            (
                NxMz_section_flipped,
                Fields_flipped,
                Bounding_flipped,
            ) = self.build2dInteractionDomainSLS(nbPoints, negative_compression)

        for nm in NxMz_section_flipped:
            nm[1] = -nm[1]

        # Adjust bounding for flipped values
        M_min = -Bounding_flipped[3]
        M_max = -Bounding_flipped[2]

        Bounding_flipped[2] = M_min
        Bounding_flipped[3] = M_max

        N_min = min(Bounding[0], Bounding_flipped[0])
        N_max = max(Bounding[1], Bounding_flipped[1])
        M_min = min(Bounding[2], Bounding_flipped[2])
        M_max = max(Bounding[3], Bounding_flipped[3])

        for si in self.__steelItems:
            si.setOriginY(-si.getOrigin().y)

        NxMz_section_flipped.reverse()
        Fields_flipped.reverse()
        return (
            NxMz_section + NxMz_section_flipped[1:],
            Fields + Fields_flipped[1:],
            [N_min, N_max, M_min, M_max],
        )

    def getH(self) -> float:
        return (
            self.__concreteItem.getShape().vertexMaxInY()[0]
            - self.__concreteItem.getShape().vertexMinInY()[0]
        )

    def getHs(self) -> float:
        return (
            self.__concreteItem.getShape().vertexMaxInY()[0]
            - self.__concreteItem.getShape().getShapePoint("G").y
        )

    def getHi(self):
        return (
            self.__concreteItem.getShape().getShapePoint("G").y
            - self.__concreteItem.getShape().vertexMinInY()[0]
        )

    def getVertexConcrAt(self, i: int) -> Point2d:
        return self.__concreteItem.getShape().vertexAt(i)

    def getVertexConcrNb(self):
        return self.__concreteItem.getShape().vertexNb()

    def getVertexSteelAt(self, i):
        return self.__steelItems[i].getShape().getOrigin()

    def getVertexSteelNb(self):
        return len(self.__steelItems)

    def build2dInteractionDomain(self, nbPoints=100, negative_compression=True):

        if not (self.__code.codeStr() == "EC2"):
            raise Exception("Implemented now only for EC2 code !!!")

        if not self.__useSectionMaterials:
            raise Exception("Implemented now only for same material on steel !!!")

        if not isinstance(self.__concreteItem.getShape(), ShapeRect):
            raise Exception("Implemented now only for rectangular section shape !!!")

        H = (
            self.__concreteItem.getShape().vertexMaxInY()[0]
            - self.__concreteItem.getShape().vertexMinInY()[0]
        )
        c = self.getSteelBotRecover()
        Hs = (
            self.__concreteItem.getShape().vertexMaxInY()[0]
            - self.__concreteItem.getShape().getShapePoint("G").y
        )
        Hi = (
            self.__concreteItem.getShape().getShapePoint("G").y
            - self.__concreteItem.getShape().vertexMinInY()[0]
        )
        Hc = (H / self.__concreteMaterial.get_ecu()) * self.__concreteMaterial.get_ec2()

        # spazio lagrangiano con il seguente significato:
        # e1: limite acciaio di massima deformazione
        # e2: limite di sezione con calcestruzzo a rottura ed acciaio a rottura (deformazioni)
        # e3: limite di sezione con calcestruzzo decompresso
        # e4: limite di sezione con calcestruzzo tutto compresso
        e1 = self.__steelMaterial.get_esu()
        e2 = e1 + self.__concreteMaterial.get_ecu()
        ecs = e2 / (H - c) * c
        e3 = e2 + ecs + self.__steelMaterial.get_esu()
        e4 = e3 + self.__concreteMaterial.get_ec2()

        # print("H = %1.6f - c = %2.6f - e1 = %3.6f - e2 = %4.6f - e3 = %5.6f - e4 = %6.6f - ecs = %7.6f"%(H,c,e1,e2,e3,e4,ecs))
        # print("Hs = %1.6f - Hi = %2.6f"%(Hs,Hi))

        B = self.__concreteItem.getShape().w()  # type: ignore

        # Convenzioni: sforzi normali positivi se di compressione
        #              flessione positiva prodotta da sforzi normali positivi lato +y
        def sigmas(e: float, temp: float = 20.0):

            fsy = self.__steelMaterial.get_fsy()
            esu = self.__steelMaterial.get_esu()

            # gammas in case of fire
            if temp <= 20.0:
                gamma_s = self.__steelMaterial.get_gammas()
            else:
                gamma_s = self.__steelMaterial.get_gammas_fire()

            #
            # reduction resistence factor in case of fire
            #
            kappa_s = 1.0
            if temp > 20.0:
                if self.__code.codeStr() in [
                    "EC2",
                    "EC2:ITA",
                    "NTC2008",
                    "NTC2018",
                    "CIRC2008",
                    "CIRC2018",
                ]:
                    if e < -0.02:
                        if (
                            self.__steelMaterial.get_shapingType()
                            == fireCode.SteelShapingType.COLD_ROLLED
                        ):
                            kappa_s = fireCode.kappa_s(
                                temp, fireCode.SteelShapingType.COLD_ROLLED
                            )
                        elif (
                            self.__steelMaterial.get_shapingType()
                            == fireCode.SteelShapingType.HOT_ROLLED
                        ):
                            kappa_s = fireCode.kappa_s(
                                temp, fireCode.SteelShapingType.HOT_ROLLED
                            )
                    else:
                        kappa_s = fireCode.kappa_s(
                            temp, stress=fireCode.StressType.COMPRESSION
                        )
                else:
                    raise Exception(
                        f"Code {self.__code.codeStr()} has not hot options !!!"
                    )

            fsd = kappa_s * fsy / gamma_s

            Es = self.__steelMaterial.get_Es()
            esd = fsd / Es

            if e >= -esd and e <= +esd:
                sigma = -Es * e
                field = "e"
            elif e < -esd and e >= -esu:
                sigma = +fsd
                field = "y"
            elif e > esd and e <= esu:
                sigma = -fsd
                field = "y"
            else:
                raise Exception("sigmas(%1.6e) deformation value error !!!" % e)

            return sigma, field, kappa_s

        def es(s, y, e1, e2, e3, e4, H, c, Hs, Hi, Hc) -> Tuple[float, float]:

            if s >= 0.0 and s < e1:
                # print("CAMPO 1")
                es = +self.__steelMaterial.get_esu() - (s / (H - c)) * (y + Hi - c)
                return es, 1.0
            elif s >= e1 and s < e2:
                # print("CAMPO 2")
                es = +self.__steelMaterial.get_esu() - (s / (H - c)) * (y + Hi - c)
                return es, 2.0
            elif s >= e2 and s < e3:
                # print("CAMPO 3")
                s1 = e3 - s + self.__concreteMaterial.get_ecu()
                es = ((s1) / H) * (Hs - y) - self.__concreteMaterial.get_ecu()
                return es, 3.0
            elif s >= e3 and s <= e4:
                # print("CAMPO 4")
                es = -(
                    ((e4 - s) / Hc) * (y + Hi - Hc) + self.__concreteMaterial.get_ec2()
                )
                return es, 4.0
            else:
                raise Exception(
                    "The value of s is <%1.4f> out of range [%2.6f,%3.6f]!!!"
                    % (s, 0, e4)
                )

        def xis(s, e1, e2, e3, e4, H, c, Hs, Hi, Hc):
            if s >= 0.0 and s < e1:
                if s == 0.0:
                    xi = -float("Inf")
                else:
                    xi = -(((H - c) / s) * (self.__steelMaterial.get_esu() - s))
            elif s >= e1 and s < e2:
                xi = -((H - c) / s) * (self.__steelMaterial.get_esu() - s)
            elif s >= e2 and s < e3:
                s1 = e3 - s + self.__concreteMaterial.get_ecu()
                xi = (self.__concreteMaterial.get_ecu() / s1) * H
            elif s >= e3 and s <= e4:
                if s == e4:
                    xi = +float("Inf")
                else:
                    xi = ((s - e3) / (e4 - s)) * Hc + H
            else:
                raise Exception(
                    "The value of s is <%1.4f> out of range [%2.6f,%3.6f]!!!"
                    % (s, 0, e4)
                )

            return xi

        def Ncs(s, e1, e2, e3, e4, H, c, Hs, Hi, Hc, hotSection: bool = False):
            lamda = self.__concreteMaterial.get_lambda()
            eta = self.__concreteMaterial.get_eta()

            xi = xis(s, e1, e2, e3, e4, H, c, Hs, Hi, Hc)

            if hotSection:
                gammac = self.__concreteMaterial.get_gammac_fire()
                alphacc = self.__concreteMaterial.get_alphacc_fire()
            else:
                gammac = self.__concreteMaterial.get_gammac()
                alphacc = self.__concreteMaterial.get_alphacc()

            if s >= 0.0 and s < e1:
                Ncs = 0.0
                ycs = 0.0
                return Ncs, ycs

            elif s >= e1 and s < e2:
                Ncs = (
                    lamda
                    * eta
                    * (self.__concreteMaterial.get_fck() * alphacc / gammac)
                    * xi
                )
                ycs = Hs - lamda * xi / 2
                return Ncs, ycs

            elif s >= e2 and s < e3:
                Ncs = (
                    lamda
                    * eta
                    * (self.__concreteMaterial.get_fck() * alphacc / gammac)
                    * xi
                )
                ycs = Hs - lamda * xi / 2
                return Ncs, ycs

            elif s >= e3 and s <= e4:
                if lamda * xi <= H:
                    Ncs = (
                        (self.__concreteMaterial.get_fck() * eta * alphacc / gammac)
                        * lamda
                        * xi
                    )
                    ycs = Hs - lamda * xi / 2
                else:
                    Ncs = (
                        self.__concreteMaterial.get_fck() * eta * alphacc / gammac
                    ) * H
                    ycs = Hs - H / 2
                return Ncs, ycs

            else:
                raise Exception(
                    "The value of s is <%1.4f> out of range [%2.6f,%3.6f]!!!"
                    % (s, 0, e4)
                )

        linearSpace = np.linspace(0, e4, nbPoints)

        NxMz_section = []
        ArrayField = []

        N_min = 0.0
        N_max = 0.0
        M_min = 0.0
        M_max = 0.0

        if True is negative_compression:
            nc = -1.0
        else:
            nc = 1.0

        ybottom = self.findLowSteelItem().getOrigin().y
        field = -1.0
        hotSection = len(self.__steelItemsTemperatures) > 0
        self.__steelItems_kappa_s = [1.0] * len(self.__steelItemsTemperatures)
        for i_s in linearSpace:
            N_steel = 0.0
            M_steel = 0.0
            steel_yelding = True
            for idx, si in enumerate(self.__steelItems):
                y_steel = si.getOrigin().y
                Ai = si.getArea()
                e, field = es(i_s, y_steel, e1, e2, e3, e4, H, c, Hs, Hi, Hc)
                if hotSection:
                    steelValues = sigmas(e, self.__steelItemsTemperatures[idx])
                    N_steel_si = (
                        steelValues[0]
                        * Ai
                        * self.__steelItemsWeights[idx]
                        * self.__steelItemsAlphaWeight
                        * nc
                    )
                    self.__steelItems_kappa_s[idx] = steelValues[2]
                else:
                    N_steel_si = (
                        sigmas(e)[0]
                        * Ai
                        * self.__steelItemsWeights[idx]
                        * self.__steelItemsAlphaWeight
                        * nc
                    )

                if (sigmas(e)[1] == "e") and (ybottom == y_steel) and steel_yelding:
                    steel_yelding = False

                N_steel = N_steel + N_steel_si
                M_steel_si = N_steel_si * y_steel * nc
                M_steel = M_steel + M_steel_si

            if len(self.__steelItemsTemperatures) > 0:
                N_concrete = (
                    Ncs(i_s, e1, e2, e3, e4, H, c, Hs, Hi, Hc, True)[0] * B * nc
                )
                y_concrete = Ncs(i_s, e1, e2, e3, e4, H, c, Hs, Hi, Hc, True)[1]
            else:
                N_concrete = (
                    Ncs(i_s, e1, e2, e3, e4, H, c, Hs, Hi, Hc, False)[0] * B * nc
                )
                y_concrete = Ncs(i_s, e1, e2, e3, e4, H, c, Hs, Hi, Hc, False)[1]

            M_concrete = N_concrete * y_concrete * nc

            N_tot = N_concrete + N_steel
            M_tot = M_concrete + M_steel

            if N_tot < N_min:
                N_min = N_tot
            if N_tot > N_max:
                N_max = N_tot
            if M_tot < M_min:
                M_min = M_tot
            if M_tot > M_max:
                M_max = M_tot

            # remove duplicates
            if len(NxMz_section) >= 1:
                if (
                    NxMz_section[len(NxMz_section) - 1][0] != N_tot
                    and NxMz_section[len(NxMz_section) - 1][1] != M_tot
                ):
                    NxMz_section.append([N_tot, M_tot])
                    # ArrayField.append(field)
                    if not steel_yelding:
                        ArrayField.append(field + 10)
                    else:
                        ArrayField.append(field)
            else:
                NxMz_section.append([N_tot, M_tot])
                # ArrayField.append(field)
                if not steel_yelding:
                    ArrayField.append(field + 10)
                else:
                    ArrayField.append(field)

        return NxMz_section, ArrayField, [N_min, N_max, M_min, M_max]

    def build2dInteractionDomainSLS(self, nbPoints=100, negative_compression=False):

        if not (self.__code.codeStr() == "EC2"):
            raise Exception("Implemented now only for EC2 code !!!")

        if not self.__useSectionMaterials:
            raise Exception("Implemented now only for same material on steel !!!")

        if not isinstance(self.__concreteItem.getShape(), ShapeRect):
            raise Exception("Implemented now only for rectangular section shape !!!")

        es_max = self.__steelMaterial.get_sigmas_max_c() / self.__steelMaterial.get_Es()
        ec_max = (
            self.__concreteMaterial.get_sigmac_max_c()
            / self.__steelMaterial.get_Es()
            * self.__homogenization
        )

        H = (
            self.__concreteItem.getShape().vertexMaxInY()[0]
            - self.__concreteItem.getShape().vertexMinInY()[0]
        )
        c = self.getSteelBotRecover()

        e1 = es_max
        e2 = ec_max + es_max
        ecs = (es_max + ec_max) / (H - c) * c
        e3 = ecs + 2 * es_max + ec_max
        e4 = ecs + 2 * (ec_max + es_max)

        Hs = (
            self.__concreteItem.getShape().vertexMaxInY()[0]
            - self.__concreteItem.getShape().getShapePoint("G").y
        )
        Hi = (
            self.__concreteItem.getShape().getShapePoint("G").y
            - self.__concreteItem.getShape().vertexMinInY()[0]
        )
        Hc = (H / self.__concreteMaterial.get_ecu()) * self.__concreteMaterial.get_ec2()

        B = self.__concreteItem.getShape().w()  # type: ignore

        # Convenzioni: sforzi normali positivi se di compressione
        #              flessione positiva prodotta da sforzi normali positivi lato +y
        def sigmas(e):

            Es = self.__steelMaterial.get_Es()
            sigma = +Es * e
            return sigma, "e"

        def es(s, y, e1, e2, e3, e4, H, c, Hs, Hi, Hc):
            di = Hi - c
            d = H - c
            if s >= 0.0 and s < e1:
                # print("CAMPO 1")
                es = es_max - (y + di) * s / d
                return -es, 1.0
            elif s >= e1 and s < e2:
                # print("CAMPO 2")
                es = +es_max - (y + di) * s / d
                return -es, 2.0
            elif s >= e2 and s < e3:
                # print("CAMPO 3")
                s1 = e4 - s
                es = +ec_max - s1 / H * (Hs - y)
                return es, 3.0
            elif s >= e3 and s <= e4:
                # print("CAMPO 4")
                s1 = e4 - s
                es = ec_max - s1 / H * (Hs - y)
                return es, 4.0
            else:
                raise Exception(
                    "The value of s is <%1.4f> out of range [%2.6f,%3.6f]!!!"
                    % (s, 0, e4)
                )

        def xis(s, e1, e2, e3, e4, H, c, Hs, Hi, Hc):

            d = H - c
            s1 = e4 - s
            if s >= 0.0 and s < e1:
                if s == 0.0:
                    xi = -float("Inf")
                else:
                    xi = -d * (es_max / s - 1)
            elif s >= e1 and s < e2:
                xi = (s - es_max) / s * d
            elif s >= e2 and s < e3:
                xi = H / s1 * ec_max
            elif s >= e3 and s <= e4:
                if s == e4:  # print("CAMPO 4")self.__concreteMaterial.get_ecu()
                    xi = +float("Inf")
                else:
                    xi = H / s1 * ec_max
            else:
                raise Exception(
                    "The value of s is <%1.4f> out of range [%2.6f,%3.6f]!!!"
                    % (s, 0, e4)
                )

            # print("xi = %1.3f"%xi)
            return xi

        def Ncs(s, e1, e2, e3, e4, H, c, Hs, Hi, Hc):

            Es = self.__steelMaterial.get_Es()
            xi = xis(s, e1, e2, e3, e4, H, c, Hs, Hi, Hc)
            n = self.__homogenization
            s1 = e4 - s

            if s >= 0.0 and s < e1:
                Ncs = 0.0
                ycs = 0.0
                return Ncs, ycs

            elif s >= e1 and s < e2:
                Ncs = xi / 2 * Es / n * (s - es_max)
                ycs = Hs - xi / 3
                return Ncs, ycs

            elif s >= e2 and s < e3:
                Ncs = xi / 2 * self.__concreteMaterial.get_sigmac_max_c()
                ycs = Hs - xi / 3
                return Ncs, ycs

            elif s >= e3 and s <= e4:
                Ncs = (
                    (2 * self.__concreteMaterial.get_sigmac_max_c() - s1 * Es / n)
                    * H
                    / 2
                )
                ycs = H * (3 * ec_max - s1) / (3 * (2 * ec_max - s1)) - Hi

                return Ncs, ycs
            else:
                raise Exception(
                    "The value of s is <%1.4f> out of range [%2.6f,%3.6f]!!!"
                    % (s, 0, e4)
                )

        linearSpace = np.linspace(0, e4, nbPoints)

        NxMz_section = []
        ArrayField = []

        N_min = 0.0
        N_max = 0.0
        M_min = 0.0
        M_max = 0.0

        ybottom = self.findLowSteelItem().getOrigin().y

        if True is negative_compression:
            nc = -1.0
        else:
            nc = 1.0

        for i_s in linearSpace:
            N_steel = 0.0
            M_steel = 0.0
            steel_yelding = True
            field = -1.0
            for idx, si in enumerate(self.__steelItems):
                y_steel = si.getOrigin().y
                Ai = si.getArea()
                e, field = es(i_s, y_steel, e1, e2, e3, e4, H, c, Hs, Hi, Hc)
                N_steel_si = (
                    sigmas(e)[0]
                    * Ai
                    * self.__steelItemsWeights[idx]
                    * self.__steelItemsAlphaWeight
                    * nc
                )

                if (sigmas(e)[1] == "e") and (ybottom == y_steel) and steel_yelding:
                    steel_yelding = False

                N_steel = (N_steel + N_steel_si) * nc
                M_steel_si = N_steel_si * y_steel * nc
                M_steel = M_steel + M_steel_si

            N_concrete = Ncs(i_s, e1, e2, e3, e4, H, c, Hs, Hi, Hc)[0] * B * nc
            y_concrete = Ncs(i_s, e1, e2, e3, e4, H, c, Hs, Hi, Hc)[1]

            # Additional parameter for benchmark
            sigma_concrete = (
                -es(i_s, Hs, e1, e2, e3, e4, H, c, Hs, Hi, Hc)[0]
                * self.__steelMaterial.get_Es()
                / self.__homogenization
            )
            y_steel_miny = self.findLowSteelItem().getShape().getOrigin().y
            sigma_steel = (
                -es(i_s, y_steel_miny, e1, e2, e3, e4, H, c, Hs, Hi, Hc)[0]
                * self.__steelMaterial.get_Es()
            )
            xi = xis(i_s, e1, e2, e3, e4, H, c, Hs, Hi, Hc)

            M_concrete = N_concrete * y_concrete * nc

            N_tot = N_concrete + N_steel
            M_tot = M_concrete + M_steel

            if N_tot < N_min:
                N_min = N_tot
            if N_tot > N_max:
                N_max = N_tot
            if M_tot < M_min:
                M_min = M_tot
            if M_tot > M_max:
                M_max = M_tot

            # remove duplicates
            if len(NxMz_section) >= 1:
                if (
                    NxMz_section[len(NxMz_section) - 1][0] != N_tot
                    and NxMz_section[len(NxMz_section) - 1][1] != M_tot
                ):
                    NxMz_section.append(
                        [
                            N_tot,
                            M_tot,
                            N_concrete,
                            M_concrete,
                            y_concrete,
                            sigma_concrete,
                            sigma_steel,
                            xi,
                        ]
                    )
                    # ArrayField.append(field)
                    if not steel_yelding:
                        ArrayField.append(field + 10)
                    else:
                        ArrayField.append(field)
            else:
                NxMz_section.append(
                    [
                        N_tot,
                        M_tot,
                        N_concrete,
                        M_concrete,
                        y_concrete,
                        sigma_concrete,
                        sigma_steel,
                        xi,
                    ]
                )
                # ArrayField.append(field)
                if not steel_yelding:
                    ArrayField.append(field + 10)
                else:
                    ArrayField.append(field)

        return NxMz_section, ArrayField, [N_min, N_max, M_min, M_max]

    def setStirrupt(
        self,
        area: Union[float, int, None] = None,
        dist: Union[float, int, None] = None,
        angle: Union[float, int, None] = None,
    ):

        # def setStirrupt(self, area = None, dist = None, angle = None):
        if area is not None:
            self.__steelStirruptArea = area
        if dist is not None:
            self.__steelStirruptDist = dist
        if angle is not None:
            self.__steelStirruptAngle = angle

    def getStirruptArea(self) -> Union[float, int, None]:
        return self.__steelStirruptArea

    def getStirruptAngle(self) -> Union[float, int, None]:
        return self.__steelStirruptAngle

    def getStirruptDistance(self) -> Union[float, int, None]:
        return self.__steelStirruptDist

    def __str__(self):
        dispstr = "ConcreteSection Object: \n"
        dispstr = dispstr + "-----------------------  \n"
        dispstr = dispstr + "  Ids = " + str(self.__ids) + "\n"
        dispstr = dispstr + "Descr = " + str(self.__descr) + "\n"
        dispstr = (
            dispstr
            + "Steel safety factor for    ULS = "
            + str(self.__steel_safetyFactorForULS)
            + "\n"
        )
        dispstr = (
            dispstr
            + "Concrete safety factor for ULS = "
            + str(self.__cls_safetyFactorForULS)
            + "\n"
        )
        dispstr = (
            dispstr
            + "Concrete Max Stress for    ELS = "
            + str(self.__concreteMaterial.get_sigmac_max_c())
            + "\n"
        )
        dispstr = (
            dispstr
            + "Steel Max Stress for       ELS = "
            + str(self.__steelMaterial.get_sigmas_max_c())
            + "\n"
        )
        dispstr = (
            dispstr
            + "Homogeneization for        ELS = "
            + str(self.__homogenization)
            + "\n"
        )
        dispstr = dispstr + "---------------------\n"
        dispstr = dispstr + "Materials embedded-->\n"
        dispstr = dispstr + "---------------------\n"
        dispstr = dispstr + str(self.__concreteMaterial).replace("\n", "\n  | ")
        dispstr = dispstr + str(self.__steelMaterial).replace("\n", "\n  | ")
        dispstr = dispstr + "-----------------\n"
        dispstr = dispstr + "Items embedded-->\n"
        dispstr = dispstr + "-----------------\n"
        if self.__concreteItem is not None:
            dispstr = dispstr + str(self.__concreteItem).replace("\n", "\n  | ")
        else:
            dispstr = dispstr + "Steel without concrete :-)\n"

        if self.__steelItems is not None:
            for i in self.__steelItems:
                dispstr = dispstr + str(i).replace("\n", "\n  | ")
        else:
            dispstr = dispstr + "Without steel :-)"
        return dispstr


class ConcreteSectionModel(BaseModel):
    descr: str = ""
    id: int = -1
    shape: RectangularShape
    concreteMat: ConcreteModel
    steelMat: SteelModel
    disposerOnLine: Optional[List[SteelDisposerOnLine]] = None
    disposerSingle: Optional[List[SteelDisposerSingle]] = None
    stirrup: Optional[SteelDisposerStirrup] = None
    stirrupSingleLeg: Optional[List[SteelDisposerStirrupSingleLeg]] = None



from typing import Any, List, Union

from pycivil.EXAUtils.EXAExceptions import EXAExceptions as ex


def validateFloat(
    var: List[Any],
    typeStr: List[Any],
    leftLimit: Union[float, None] = None,
    leftLimitEq: Union[float, None] = None,
    rightLimit: Union[float, None] = None,
    rightLimitEq: Union[float, None] = None,
) -> None:
    if not isinstance(var, list):
        raise ex("001", "var must be a list", type(var))

    if not isinstance(typeStr, list):
        raise ex("001", "typeStr must be a list of str type", type(typeStr))

    for v in var:
        if type(v) not in typeStr:
            raise ex("002", "var must be a typeStr", type(v))

        if all(
            [
                leftLimit is None,
                leftLimitEq is None,
                rightLimit is None,
                rightLimitEq is None,
            ]
        ):
            return
        else:
            if leftLimit is not None and leftLimitEq is None:
                if v <= leftLimit:
                    raise ex(
                        "003",
                        f"var must be greater than {leftLimit:.3f}",
                        type(v),
                    )
            if leftLimitEq is not None and leftLimit is None:
                if v < leftLimit:
                    raise ex(
                        "004",
                        f"var must be greater or equal than {leftLimitEq:.3f}",
                        type(v),
                    )
            if rightLimit is not None and rightLimitEq is None:
                if v <= leftLimit:
                    raise ex(
                        "005",
                        f"var must be less than {rightLimit:.3f}",
                        type(v),
                    )
            if rightLimitEq is not None and rightLimit is None:
                if v < leftLimit:
                    raise ex(
                        "006",
                        f"var must be less or equal than {rightLimitEq:.3f}",
                        type(v),
                    )



"""Module of rules to build EC2 (Eurocode 2) compliant objects.

An example of compliant objects is reinforced concrete section for
plate elements with minimum reinforcement.

Classes list:

    1. PlateMinimumAreaInput
    2. PlateMinimumAreaOutput
    3. PlateMinimumAreaLogs
    4. PlateMinimumArea
    5. SolverPlateMinRebar
"""
from enum import Enum
from typing import Any, Union

from pydantic import BaseModel, field_validator, model_validator
from typing_extensions import Self

from pycivil.EXAStructural.lawcodes.codeEC211 import (
    rcSectionBeamAreaMin,
    rcSectionPlateShearAreaMin,
    rcSectionPlateShearLegsMax,
    rcSectionPlateStepMax,
    rcSectionPlateStirrupStepsMax,
    rebarArea,
    rebarNumber,
)
from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.materials import Concrete, ConcreteSteel
from pycivil.EXAUtils.solver import Solver


class PlateMinimumAreaInput(BaseModel):
    """Input model with datas for minimum area calculation

    ...
    Attributes:
        elementDescr (str, optional): element description.
         Default is "Nome elemento"
        keyCode (str, optional): key code string.
         Default is "NTC2018"
        steelClass (str, optional): key class for steel.
         Default is "B450C"
        concreteClass (str, optional): key class for concrete.
         Default is "C20/25"
        hEl (float, optional): height for element.
         Default is 600.0
        rebarD (float, optional): rebar diameter for main reinforcement
         Default is 12
        cover (float, optional): float.
         Default is 52.0
        rebarDSec (float, optional): float.
         Default is 12
        coverSec (float, optional): float.
         Default is 40.0
        stirrupD (float, optional): float.
         Default is 8
        nbLegDirX (float, optional): float.
         Default is 2
    """

    elementDescr: str = "Nome elemento"
    keyCode: str = "NTC2018"
    steelClass: str = "B450C"
    concreteClass: str = "C20/25"
    hEl: float = 600.0
    rebarD: float = 12
    cover: float = 52.0
    rebarDSec: float = 12
    coverSec: float = 40.0
    stirrupD: float = 8
    nbLegDirX: float = 2

    class ConfigDict:
        validate_assignment = True

    @field_validator("elementDescr", "keyCode", "steelClass", "concreteClass", mode="before")
    @classmethod
    def must_be_string(cls, v: Any) -> Any:
        assert isinstance(v, str)
        return v

    @field_validator(
        "hEl", "rebarD", "cover", "rebarDSec", "coverSec", "stirrupD", "nbLegDirX", mode="before"
    )
    @classmethod
    def greater_than_zero(cls, v: Any) -> Any:
        assert isinstance(v, float) or isinstance(v, int)
        assert v > 0
        return v

    @model_validator(mode="after")
    def rebar_in_concrete(self) -> Self:
        if (
            self.hEl - 2 * max(self.rebarD + self.cover, self.rebarDSec + self.coverSec)
            < 0
        ):
            raise ValueError("Concrete height")
        if not (self.coverSec + self.rebarDSec <= self.cover) and not (
            self.coverSec >= self.cover + self.rebarD
        ):
            raise ValueError("Steel collision")
        return self


class PlateMinimumAreaOutput(BaseModel):
    """Output model for minimum area calculation

    ...
    Attributes:
        rebarNumberNormal (float, optional): Main reinforcement minimum
         number with big normal force. Default is  0.0
        rebarNumberMoment (float, optional): Main reinforcement minimum
         number with big moment. Default is  0.0
        rebarNumberNormalSec (float, optional): Secondary reinforcement
         minimum number with big normal force. Default is  0.0
        rebarNumberMomentSec (float, optional): Secondary reinforcement
         minimum number with big moment. Default is  0.0
        stirrupStepMin (float, optional): Stirrup minimum step
         Default is  0.0
        nbLegDirXDisposed (float, optional): Stirrup legs minimum.
         Default is  0.0
    """
    rebarNumberNormal: float = 0.0
    rebarNumberMoment: float = 0.0
    rebarNumberNormalSec: float = 0.0
    rebarNumberMomentSec: float = 0.0
    stirrupStepMin: float = 0.0
    nbLegDirXDisposed: float = 0.0


class PlateMinimumAreaLogs(BaseModel):
    cls_fck: float = 0
    cls_fctm: float = 0
    steel_fyk: float = 0
    heightUtil: float = 0
    areaUtil: float = 0
    minimumRebarAreaCrit1: float = 0
    minimumRebarAreaCrit2: float = 0
    minimumRebarArea: float = 0
    distMaxRebar: float = 0
    distMaxRebarMaxLoad: float = 0
    disposedRebarArea: float = 0
    disposedRebarNumber: float = 0
    disposedRebarAreaMaxLoad: float = 0
    disposedRebarNumberMaxLoad: float = 0
    heightUtilSec: float = 0
    areaUtilSec: float = 0
    minimumRebarAreaCrit1Sec: float = 0
    minimumRebarAreaCrit2Sec: float = 0
    minimumRebarAreaSec: float = 0
    distMaxRebarSec: float = 0
    distMaxRebarMaxLoadSec: float = 0
    disposedRebarAreaSec: float = 0
    disposedRebarNumberSec: float = 0
    disposedRebarAreaMaxLoadSec: float = 0
    disposedRebarNumberMaxLoadSec: float = 0
    minimumRebarAreaForElementLenght: float = 0
    maxStepTrasv: float = 0
    legsNumber: float = 0
    legsNumberTrasv: float = 0
    maxStepLongCrit1: float = 0
    maxStepLongCrit2: float = 0
    maxStep: float = 0
    rebarAreaForElementLenght: float = 0


class PlateMinimumArea(BaseModel):
    """Model for minimum area calculation with SolverPlateMinRebar()

    ...
    Attributes:
        inputData (PlateMinimumAreaInput, optional): Input model with
         datas for minimum area calculation. Default is
         PlateMinimumAreaInput()
        logsData (PlateMinimumAreaLogs, optional): Log model. Default
         is PlateMinimumAreaLogs()
        outputData (PlateMinimumAreaOutput, optional): Output model
         for minimum area calculation. Default is
         PlateMinimumAreaOutput()
        started (bool, optional): Say if solver is started. Default is
         False
        success (bool, optional): Say if solver exit with success.
         Default is False
    """
    inputData: PlateMinimumAreaInput = PlateMinimumAreaInput()
    logsData: PlateMinimumAreaLogs = PlateMinimumAreaLogs()
    outputData: PlateMinimumAreaOutput = PlateMinimumAreaOutput()
    started: bool = False
    success: bool = False


class SolverPlateMinRebar(Solver):
    """Solver for minimum area calculation

    ...
    """
    def __init__(self, inputModel: PlateMinimumAreaInput):
        """Initializes the instance

        Args:
            inputModel (PlateMinimumAreaInput):
        """
        if not isinstance(inputModel, PlateMinimumAreaInput):
            raise ValueError
        self.__in = inputModel
        super().__init__(modelInput=inputModel)
        self.__logs = PlateMinimumAreaLogs()
        super()._setModelLogs(model=self.__logs)
        self.__out = PlateMinimumAreaOutput()
        super()._setModelOutput(model=self.__out)

    def run(self, opt: Union[Enum, None] = None, **kwargs) -> bool:
        """Run the solver

        The solver using input data PlateMinimumAreaInput class, calculates
        minimal area for longitudinal reinforcement and stittups.

        Args:
            opt (): no option for this solver
            **kwargs (): none kwargs for this

        Returns:
            True if exit with success
        """
        # Materials
        mat_steel = ConcreteSteel()
        mat_steel.setByCode(
            codeObj=Code(self.__in.keyCode), catstr=self.__in.steelClass
        )
        self.__logs.steel_fyk = mat_steel.get_fsy()
        mat_concrete = Concrete()
        mat_concrete.setByCode(
            codeObj=Code(self.__in.keyCode), catstr=self.__in.concreteClass
        )
        self.__logs.cls_fck = mat_concrete.get_fck()
        self.__logs.cls_fctm = mat_concrete.get_fctm()

        # Geometry for main rebar
        self.__logs.heightUtil = self.__in.hEl - self.__in.cover - self.__in.rebarD / 2

        self.__logs.minimumRebarAreaCrit1 = rcSectionBeamAreaMin(
            self.__logs.cls_fck, self.__logs.steel_fyk, self.__logs.heightUtil, "c1"
        )
        self.__logs.minimumRebarAreaCrit2 = rcSectionBeamAreaMin(
            self.__logs.cls_fck, self.__logs.steel_fyk, self.__logs.heightUtil, "c2"
        )
        self.__logs.minimumRebarArea = rcSectionBeamAreaMin(
            self.__logs.cls_fck,
            self.__logs.steel_fyk,
            self.__logs.heightUtil,
        )
        self.__logs.areaUtil = 1000 * self.__logs.heightUtil

        rebarNb = rebarNumber(self.__logs.minimumRebarArea, self.__in.rebarD)[0]
        step = 1000 / rebarNb

        stepMaxNormal = rcSectionPlateStepMax(self.__in.hEl, "main", "normal")
        self.__logs.distMaxRebar = stepMaxNormal

        stepMaxMoment = rcSectionPlateStepMax(self.__in.hEl, "main", "maxbending")
        self.__logs.distMaxRebarMaxLoad = stepMaxMoment

        rebarNumberNormal = rebarNb
        # if step > stepMaxNormal:
        #     for _i in range(0, 1000):
        #         rebarNumberNormal += 1
        #         if 1000 / rebarNumberNormal <= stepMaxNormal:
        #             break

        if step > stepMaxNormal:
            rebarNumberNormal = 1000 / stepMaxNormal

        rebarNumberMoment = rebarNb
        # rebarNumberMoment = rebarNb
        # if step > stepMaxMoment:
        #     for _i in range(0, 1000):
        #         rebarNumberMoment += 1
        #         if 1000 / rebarNumberMoment <= stepMaxMoment:
        #             break

        if step > stepMaxMoment:
            rebarNumberMoment = 1000 / stepMaxMoment

        self.__logs.disposedRebarNumber = rebarNumberNormal
        self.__logs.disposedRebarArea = rebarNumberNormal * rebarArea(self.__in.rebarD)

        self.__logs.disposedRebarNumberMaxLoad = rebarNumberMoment
        self.__logs.disposedRebarAreaMaxLoad = rebarNumberMoment * rebarArea(
            self.__in.rebarD
        )

        self.__out.rebarNumberNormal = self.__logs.disposedRebarNumber
        self.__out.rebarNumberMoment = self.__logs.disposedRebarNumberMaxLoad

        # Geometry for secondary rebar
        self.__logs.heightUtilSec = (
            self.__in.hEl - self.__in.coverSec - self.__in.rebarDSec / 2
        )

        self.__logs.minimumRebarAreaCrit1Sec = rcSectionBeamAreaMin(
            self.__logs.cls_fck, self.__logs.steel_fyk, self.__logs.heightUtilSec, "c1"
        )
        self.__logs.minimumRebarAreaCrit2Sec = rcSectionBeamAreaMin(
            self.__logs.cls_fck, self.__logs.steel_fyk, self.__logs.heightUtilSec, "c2"
        )
        self.__logs.minimumRebarAreaSec = rcSectionBeamAreaMin(
            self.__logs.cls_fck,
            self.__logs.steel_fyk,
            self.__logs.heightUtilSec,
        )
        self.__logs.areaUtilSec = 1000 * self.__logs.heightUtilSec

        rebarNbSec = rebarNumber(self.__logs.minimumRebarAreaSec, self.__in.rebarDSec)[
            0
        ]
        stepSec = 1000 / rebarNbSec

        stepMaxNormalSec = rcSectionPlateStepMax(self.__in.hEl, "secondary", "normal")
        self.__logs.distMaxRebarSec = stepMaxNormalSec

        stepMaxMomentSec = rcSectionPlateStepMax(
            self.__in.hEl, "secondary", "maxbending"
        )
        self.__logs.distMaxRebarMaxLoadSec = stepMaxMomentSec

        rebarNumberNormalSec = rebarNbSec
        if stepSec > stepMaxNormalSec:
            for _i in range(0, 1000):
                rebarNumberNormalSec += 1
                if 1000 / rebarNumberNormalSec <= stepMaxNormalSec:
                    break

        rebarNumberMomentSec = rebarNbSec
        if stepSec > stepMaxMomentSec:
            for _i in range(0, 1000):
                rebarNumberMomentSec += 1
                if 1000 / rebarNumberMomentSec <= stepMaxMomentSec:
                    break

        self.__logs.disposedRebarNumberSec = rebarNumberNormalSec
        self.__logs.disposedRebarAreaSec = rebarNumberNormalSec * rebarArea(
            self.__in.rebarDSec
        )

        self.__logs.disposedRebarNumberMaxLoadSec = rebarNumberMomentSec
        self.__logs.disposedRebarAreaMaxLoadSec = rebarNumberMomentSec * rebarArea(
            self.__in.rebarDSec
        )

        self.__out.rebarNumberNormalSec = self.__logs.disposedRebarNumberSec
        self.__out.rebarNumberMomentSec = self.__logs.disposedRebarNumberMaxLoadSec

        # Shear rebar
        self.__logs.minimumRebarAreaForElementLenght = rcSectionPlateShearAreaMin(
            self.__logs.cls_fck, self.__logs.steel_fyk
        )
        self.__logs.legsNumber = rebarNumber(
            self.__logs.minimumRebarAreaForElementLenght, self.__in.stirrupD
        )[0]

        self.__logs.maxStepTrasv = rcSectionPlateShearLegsMax(
            min(self.__logs.heightUtil, self.__logs.heightUtilSec)
        )

        nbLegDirXDisposed = self.__in.nbLegDirX

        # if 1000 / nbLegDirXDisposed > self.__logs.maxStepTrasv:
        #     for _i in range(0, 100):
        #         nbLegDirXDisposed += 1
        #         if 1000 / nbLegDirXDisposed < self.__logs.maxStepTrasv:
        #             break

        if 1000 / nbLegDirXDisposed > self.__logs.maxStepTrasv:
            nbLegDirXDisposed = 1000 / self.__logs.maxStepTrasv

        self.__logs.maxStepLongCrit1 = nbLegDirXDisposed / self.__logs.legsNumber * 1000
        self.__logs.maxStepLongCrit2 = rcSectionPlateStirrupStepsMax(
            min(self.__logs.heightUtil, self.__logs.heightUtilSec)
        )
        self.__logs.maxStep = min(
            [self.__logs.maxStepLongCrit1, self.__logs.maxStepLongCrit2]
        )

        self.__out.stirrupStepMin = self.__logs.maxStep
        self.__out.nbLegDirXDisposed = nbLegDirXDisposed
        self.__logs.legsNumberTrasv = nbLegDirXDisposed
        self.__logs.rebarAreaForElementLenght = (
            nbLegDirXDisposed
            * rebarArea(self.__in.stirrupD)
            * 1000
            / self.__out.stirrupStepMin
        )

        return True



import math

from pycivil.EXAStructural import templateRCRect as est

# Setting new instance of section with
# id = 1 and name = "First Section"
section = est.RCTemplRectEC2(1, "First Section")
section.setElementType("beam")

# Setting dimension
section.setDimH(500.0)
section.setDimW(300.0)

# Setting materials
section.setMaterials("C32/40", "B450C")

# Adding Tension point Area
pt = est.eg.Point2d(+2.5 * 1e3 * 1e3, 2.93 * 1e2 * 1e6)
section.addTensionPoint2d(N=pt.x, M=pt.y)
# section can have many tension points then we need choosing the current
section.setCurrentIdxTensionPoint2d(0)

# Finding area max and min with current tension point
print("Values after tension point")
print("Steel Area max = %1.3f" % section.calSteelAreaMax())
print("Steel Area min = %1.3f" % section.calSteelAreaMin())

AsteelMax = section.calSteelAreaMax()
AsteelMin = section.calSteelAreaMin()
AsteelMed = (AsteelMax + AsteelMin) / 2

# 'MB' means medium bottom area
# 'MT' means medium top area
section.addSteelArea("MB", 40.0, AsteelMed / 2)
section.addSteelArea("MT", 40.0, AsteelMed / 2)

# Building points set of interaction diagramm with min and max
areaFactor0 = 0
minPointsCloud = section.interactionDomainBuild2d(
    nbPoints=100, alpha=section.alphaPhi(AsteelMin, AsteelMax, areaFactor0)
)
areaFactor1 = 1
maxPointsCloud, bounding = section.interactionDomainBuild2d(
    nbPoints=100,
    alpha=section.alphaPhi(AsteelMin, AsteelMax, areaFactor1),
    bounding=True,
)

print("Bounding of MAXIMUM domain")
print(bounding)

brox = bounding[1] - bounding[0]
broy = bounding[3] - bounding[2]

linesOfMinimalPoints = []

contained, pintersect, intfactor0, other = minPointsCloud.contains(
    pt.x, pt.y, rayFromCenter=True, ro=(brox, broy)
)

print("Intersection factor")
print(intfactor0)
section.addTensionPoints2d(pintersect)
if contained:
    raise Exception("Pt - Containted in MIN !!!")
else:
    print("[500.0,191.2] - Not containted in MIN !!!")
    # print("Minimal distance: %1.6e"%minDistance)
    if len(pintersect) == 2:
        linesOfMinimalPoints.append([pt, pintersect[0]])
        linesOfMinimalPoints.append([pt, pintersect[1]])

contained, pintersect, intfactor1, other = maxPointsCloud.contains(
    pt.x, pt.y, rayFromCenter=True, ro=(brox, broy)
)

print("Intersection factor")
print(intfactor1)
section.addTensionPoints2d(pintersect)
if contained:
    print("[+2.5*1E+3*1E+3,2.93*1E+2*1E+6] - Containted in MAX !!!")
    # print("Minimal distance: %1.6e"%minDistance)
    if len(pintersect) == 2:
        linesOfMinimalPoints.append([pt, pintersect[0]])
        linesOfMinimalPoints.append([pt, pintersect[1]])
else:
    raise Exception("Pt - Not containted in MAX !!!")

Area_n = areaFactor1
Area_nm1 = areaFactor0
Fx_n = intfactor1 - 1.0
Fx_nm1 = intfactor0 - 1.0

toll = 1e-5
maxIt = 100
for i in range(1, maxIt):
    print("Step i= %1.i" % i)

    Area_np1 = Area_n - ((Area_n - Area_nm1) / (Fx_n - Fx_nm1)) * Fx_n
    print("Area_np1= %1.6e" % Area_np1)

    Area_nm1 = Area_n
    Area_n = Area_np1

    Fx_nm1 = Fx_n

    pointsCloud = section.interactionDomainBuild2d(
        nbPoints=100, alpha=section.alphaPhi(AsteelMin, AsteelMax, Area_n)
    )
    contained, pintersect, intfactor, other = pointsCloud.contains(
        pt.x, pt.y, rayFromCenter=True, ro=(brox, broy)
    )

    Fx_n = intfactor - 1.0
    print("toll = %1.6e" % math.sqrt(math.pow(Fx_n, 2)))

    if math.sqrt(math.pow(Fx_n, 2)) < toll:
        break


# Plot
section.interactionDomainPlot2d(lines=linesOfMinimalPoints, markers=False)



import os
import shutil
from pathlib import Path
from typing import List, Literal, Union
from uuid import uuid4

import gmsh
import numpy as np
from jinja2 import Environment, FileSystemLoader

import pycivil.EXAUtils.vtk as vtk
from pycivil.EXAGeometry.geometry import Point2d, Point3d
from pycivil.EXAStructural.lawcodes.codeEC212 import FireCurve, LimitCurve, Moisture, RMinutes
from pycivil.EXAStructural.templateRCRect import RCTemplRectEC2
from pycivil.EXAUtils import logging as logger
from code_aster_whale.tools.code_aster import AsterLauncher, ExportType, LauncherType


class RCRectEC2FireDesign:
    def __init__(
        self,
        coldSection: Union[RCTemplRectEC2, None] = None,
        workingPath: str = "",
        logLevel: Literal[0, 1, 2, 3] = 1,
        jobToken: str = "",
        codeAsterLauncher: str = "CONTAINER",
        codeAsterTemplatePath: Path = Path(__file__).parent.parent.parent
        / "res"
        / "codeaster",
        codeAsterContainerName: str = "",
        paraviewTemplatesPath: str = "",
    ):
        self.__codeAsterLauncher = codeAsterLauncher
        self.__codeAsterContainerName = codeAsterContainerName
        self.__codeAsterTemplatePath = codeAsterTemplatePath
        self.__paraviewTemplatesPath = paraviewTemplatesPath
        self.__section = coldSection
        self.__logLevel = logLevel
        self.__workingPath = os.getcwd()
        self.__fileName_mesh = "thermal-rect"
        self.__ext_mesh = "med"
        self.__ext_mesh_aster = "mmed"
        self.__aster_job_uuid = str(uuid4())
        self.__meshNumberInX = 15
        self.__meshNumberInY = 30
        logger.log(
            "INF",
            f"Aster job   uuid --> *{self.__aster_job_uuid}*",
            self.__logLevel,
        )
        self.__aster_rep_uuid = str(uuid4())
        logger.log(
            "INF",
            f"Aster rep   uuid --> *{self.__aster_rep_uuid}*",
            self.__logLevel,
        )
        self.__aster_study_uuid = str(uuid4())
        logger.log(
            "INF",
            f"Aster study uuid --> *{self.__aster_rep_uuid}*",
            self.__logLevel,
        )
        self.__aster_study_name = "thermal-rect"
        logger.log(
            "INF",
            f"Aster study name --> *{self.__aster_study_name}*",
            self.__logLevel,
        )
        self.__fire_moisture_curve = Moisture.v00
        self.__fire_limite_curve = LimitCurve.INF
        self.__fire_curve = FireCurve.RWS
        self.__parser = vtk.Parser()
        if self.__codeAsterLauncher == "LOCAL":
            logger.log("INF", "Run Aster as *LOCAL*", self.__logLevel)
            self.__launcher = AsterLauncher(
                launcherType=LauncherType.SINGULARITY,
                exportType=ExportType.VTK_TIME_SERIES,
                containerName=self.__codeAsterContainerName,
                paraviewTemplatesPath=self.__paraviewTemplatesPath,
            )
            self.__parser.setFileType(vtk.FileType.VTM_TYPE)

        elif self.__codeAsterLauncher == "CONTAINER":
            logger.log("INF", "Run Aster as *CONTAINER*", self.__logLevel)
            self.__launcher = AsterLauncher(
                launcherType=LauncherType.CONTAINER,
                exportType=ExportType.VTK_FROM_MEDCOUPLING,
                jobToken=jobToken,
                appName="RCRectangular",
                containerName=self.__codeAsterContainerName,
                paraviewTemplatesPath=self.__paraviewTemplatesPath,
            )
            self.__parser.setFileType(vtk.FileType.VTU_TYPE)
        else:
            logger.log(
                "ERR",
                f"Aster settings unknown --> *{self.__codeAsterLauncher}*",
                self.__logLevel,
            )

        logger.log(
            "INF",
            f"Aster settings --> *{self.__codeAsterLauncher}*",
            self.__logLevel,
        )

        self.__exposedTop = False
        self.__exposedBottom = True
        self.__exposedLeft = False
        self.__exposedRight = False
        self.__time = 180
        self.__reductionLeft = 0.0
        self.__reductionRight = 0.0
        self.__reductionTop = 0.0
        self.__reductionBottom = 0.0

        if workingPath != "":
            if os.path.exists(os.path.expanduser(workingPath)):
                self.__workingPath = workingPath
            else:
                self.__workingPath = os.getcwd()
                logger.log(
                    "ERR",
                    "Working path {} do not exists !!! Make cwd {}".format(
                        workingPath, os.getcwd()
                    ),
                    self.__logLevel,
                )

        # Trash file and dir
        self.__fileNameToTrash = []
        self.__dirsNameToTrash = []

    def __fileFullPathComm(self) -> str:
        return os.path.join(self.__workingPath, self.__aster_study_name + ".comm")

    def __fileFullPathExport(self) -> str:
        return os.path.join(self.__workingPath, self.__aster_study_name + ".export")

    def __fileFullPathRmed(self) -> str:
        return os.path.join(self.__workingPath, self.__aster_study_name + ".rmed")

    def __fileFullPathMmed(self) -> str:
        return os.path.join(
            self.__workingPath, self.__fileName_mesh + "." + self.__ext_mesh_aster
        )

    def getFileNameToTrash(self):
        return self.__fileNameToTrash + self.__launcher.getArtifactsFileName()

    def getDirsNameToTrash(self):
        return self.__dirsNameToTrash + self.__launcher.getArtifactsDirName()

    def setLogLevel(self, ll: int) -> bool:
        if 0 <= ll <= 3:
            self.__logLevel = ll
            return True
        else:
            logger.log("ERR", "Log level must be between 0 and 3 !!!")
            return False

    def setMoisture(self, moisture: Moisture):
        self.__fire_moisture_curve = moisture

    def setFireCurve(self, curve: FireCurve):
        self.__fire_curve = curve

    def setEsposedTop(self, exposed=True):
        self.__exposedTop = exposed

    def setEsposedBottom(self, exposed=True):
        self.__exposedBottom = exposed

    def setEsposedLeft(self, exposed=True):
        self.__exposedLeft = exposed

    def setEsposedRight(self, exposed=True):
        self.__exposedRight = exposed

    def setSection(self, coldSection: RCTemplRectEC2):
        self.__section = coldSection

    def getSection(self) -> Union[RCTemplRectEC2, None]:
        return self.__section

    def setTime(self, t: RMinutes = RMinutes.R180):
        self.__time = int(t.name[1 : len(t.name)])

    def getTime(self):
        return self.__time

    def getCurve(self):
        return self.__fire_curve

    def setWorkingPath(self, workingPath: str, purge: bool = False) -> bool:
        if os.path.exists(os.path.expanduser(workingPath)):
            self.__workingPath = os.path.expanduser(workingPath)
            logger.log(
                "INF",
                f"Working path *{self.__workingPath}* setted.",
                self.__logLevel,
            )
            if purge:
                listFiles = os.listdir(self.__workingPath)
                for baseName in listFiles:
                    completeName = os.path.join(self.__workingPath, baseName)
                    if os.path.isfile(completeName):
                        try:
                            os.remove(completeName)
                            logger.log(
                                "WRN",
                                f"Removed file --> {completeName}",
                                self.__logLevel,
                            )
                        except (FileNotFoundError, OSError):
                            logger.log(
                                "ERR",
                                f"Removing file --> {completeName}",
                                self.__logLevel,
                            )
                    elif os.path.isdir(os.path.join(self.__workingPath, baseName)):
                        try:
                            shutil.rmtree(completeName)
                            logger.log(
                                "WRN",
                                f"Removed dir --> {completeName}",
                                self.__logLevel,
                            )
                        except OSError:
                            logger.log(
                                "ERR",
                                f"Removing dir --> {completeName}",
                                self.__logLevel,
                            )
                        pass
                    else:
                        pass

            return True
        else:
            logger.log("ERR", "Working path {} do not exists !!!", self.__logLevel)
            return False

    def getWorkingPath(self) -> str:
        return self.__workingPath

    def buildMesh(self, ll: Literal[0, 1, 2, 3] = 3) -> bool:
        fileNameForMed = os.path.join(
            self.__workingPath, self.__fileName_mesh + "." + self.__ext_mesh
        )
        fileNameForCodeAster = self.__fileFullPathMmed()

        if os.path.exists(fileNameForCodeAster):
            logger.log("WRN", "Mesh yet exists. !!! overwrite", ll)
            # return True

        if self.__section is None:
            logger.log("ERR", "Section is None. My be use setSection() !!! quit", ll)
            return False

        height = self.__section.getDimH()
        width = self.__section.getDimW()

        # GMSH need to be initialized
        #
        gmsh.initialize()

        gmsh.model.add("EXAStruTemplate")
        gmsh.model.set_current("EXAStruTemplate")

        # Points

        p_BL = gmsh.model.occ.addPoint(
            x=-width / 2 / 1000, y=-height / 2 / 1000, z=0, meshSize=0.0, tag=1
        )
        p_BR = gmsh.model.occ.addPoint(
            x=+width / 2 / 1000, y=-height / 2 / 1000, z=0, meshSize=0.0, tag=2
        )
        p_TR = gmsh.model.occ.addPoint(
            x=+width / 2 / 1000, y=+height / 2 / 1000, z=0, meshSize=0.0, tag=3
        )
        p_TL = gmsh.model.occ.addPoint(
            x=-width / 2 / 1000, y=+height / 2 / 1000, z=0, meshSize=0.0, tag=4
        )

        # Lines
        l_BOTTOM = gmsh.model.occ.addLine(p_BL, p_BR, tag=11)
        l_RIGHT = gmsh.model.occ.addLine(p_BR, p_TR, tag=12)
        l_TOP = gmsh.model.occ.addLine(p_TR, p_TL, tag=13)
        l_LEFT = gmsh.model.occ.addLine(p_TL, p_BL, tag=14)

        # Surface
        wireId = gmsh.model.occ.addCurveLoop([l_BOTTOM, l_RIGHT, l_TOP, l_LEFT], tag=1)
        s_RECT = gmsh.model.occ.addPlaneSurface(wireTags=[wireId], tag=1)

        # GMSH need syncronize geometrycal entities with model
        #
        gmsh.model.occ.synchronize()

        # Points physical group
        p_TL_pg = gmsh.model.addPhysicalGroup(0, [p_TL])
        p_TR_pg = gmsh.model.addPhysicalGroup(0, [p_TR])
        p_BL_pg = gmsh.model.addPhysicalGroup(0, [p_BL])
        p_BR_pg = gmsh.model.addPhysicalGroup(0, [p_BR])
        gmsh.model.setPhysicalName(0, p_TL_pg, "TL")
        gmsh.model.setPhysicalName(0, p_TR_pg, "TR")
        gmsh.model.setPhysicalName(0, p_BL_pg, "BL")
        gmsh.model.setPhysicalName(0, p_BR_pg, "BR")

        # Lines physical group
        l_LEFT_pg = gmsh.model.addPhysicalGroup(1, [l_LEFT])
        l_RIGHT_pg = gmsh.model.addPhysicalGroup(1, [l_RIGHT])
        l_TOP_pg = gmsh.model.addPhysicalGroup(1, [l_TOP])
        l_BOTTOM_pg = gmsh.model.addPhysicalGroup(1, [l_BOTTOM])
        gmsh.model.setPhysicalName(1, l_LEFT_pg, "LEFT")
        gmsh.model.setPhysicalName(1, l_RIGHT_pg, "RIGHT")
        gmsh.model.setPhysicalName(1, l_TOP_pg, "TOP")
        gmsh.model.setPhysicalName(1, l_BOTTOM_pg, "BOTTOM")

        # Surface physical group
        s_RECT_pg = gmsh.model.addPhysicalGroup(2, [s_RECT])
        gmsh.model.setPhysicalName(2, s_RECT_pg, "RECT")

        mesher = gmsh.model.mesh()
        mesher.setTransfiniteCurve(l_LEFT, self.__meshNumberInY + 1)
        mesher.setTransfiniteCurve(l_RIGHT, self.__meshNumberInY + 1)
        mesher.setTransfiniteCurve(l_TOP, self.__meshNumberInX + 1)
        mesher.setTransfiniteCurve(l_BOTTOM, self.__meshNumberInX + 1)
        mesher.setTransfiniteSurface(tag=1, arrangement="left", cornerTags=[1, 2, 3, 4])

        # Comment this only for generate triangular meshes
        # mesher.setRecombine(dim = 2, tag = 1)

        mesher.generate(2)

        gmsh.write(fileNameForMed)
        # GMSH need to be initialized
        #
        gmsh.finalize()

        if os.path.exists(fileNameForCodeAster):
            logger.log(
                "INF",
                f"Mesh file *{fileNameForCodeAster}* already exist ... will be removed.",
                self.__logLevel,
            )
            os.remove(fileNameForCodeAster)

        os.rename(fileNameForMed, fileNameForCodeAster)
        if os.path.exists(fileNameForCodeAster):
            logger.log(
                "INF",
                f"Mesh file *{fileNameForCodeAster}* created.",
                self.__logLevel,
            )
            self.__fileNameToTrash.append(fileNameForCodeAster)
            return True
        else:
            logger.log(
                "ERR",
                f"Can not create mesh file *{fileNameForCodeAster}*.",
                self.__logLevel,
            )
            return False

    def buildThermalMap(self, test=False) -> bool:
        os.path.dirname(os.path.abspath(__file__))

        file_loader = FileSystemLoader(searchpath=self.__codeAsterTemplatePath)
        env = Environment(loader=file_loader)

        templateName = "template-thermal-rect.export"
        jtemplate = env.get_template(templateName)
        placeHolders_export = {
            "job_uuid": self.__aster_job_uuid,
            "rep_uuid": self.__aster_rep_uuid,
            "study_uuid": self.__aster_study_uuid,
            "study_name": self.__aster_study_name,
        }
        rendered = jtemplate.render(placeHolders_export)

        fileNameExport = self.__fileFullPathExport()
        if os.path.exists(fileNameExport):
            # logger.log('WRN','Code aster export yet exist !!! do nothing',self.__logLevel)
            logger.log(
                "WRN", "Code aster export yet exist !!! Overwrite", self.__logLevel
            )

        try:
            with open(fileNameExport, "w") as f:
                f.write(rendered)  # type: ignore
        except OSError:
            logger.log("ERR", f"Writing file *{fileNameExport}*.", self.__logLevel)
            return False
        else:
            logger.log(
                "INF",
                f"File for aster EXPORT *{fileNameExport}* created.",
                self.__logLevel,
            )
            self.__fileNameToTrash.append(fileNameExport)

        templateName = "template-thermal-rect.comm"
        jtemplate = env.get_template(templateName)
        boundary_list = (
            self.__exposedBottom * "'BOTTOM',"
            + self.__exposedTop * "'TOP',"
            + self.__exposedLeft * "'LEFT',"
            + self.__exposedRight * "'RIGHT',"
        )

        fire_curve = "FIRERWS"
        if self.__fire_curve == FireCurve.RWS:
            fire_curve = "FIRERWS"
        elif self.__fire_curve == FireCurve.ISO834:
            fire_curve = "FIREISO"
        elif self.__fire_curve == FireCurve.HC:
            fire_curve = "FIREHC"
        elif self.__fire_curve == FireCurve.HCM:
            fire_curve = "FIREHCM"
        else:
            logger.log("ERR", "Fire Curve option unknown", self.__logLevel)

        capacity_curve = "Cv00"
        if self.__fire_moisture_curve == Moisture.v00:
            capacity_curve = "Cv00"
        elif self.__fire_moisture_curve == Moisture.v15:
            capacity_curve = "Cv15"
        elif self.__fire_moisture_curve == Moisture.v30:
            capacity_curve = "Cv30"
        else:
            logger.log("ERR", "Moisture Curve option unknown", self.__logLevel)

        placeHolders_command = {
            "capacity_curve": capacity_curve,
            "fire_curve": fire_curve,
            "conduttivity": "CONDMIN",
            "boundary_list": boundary_list,
        }
        rendered = jtemplate.render(placeHolders_command)

        fileNameComm = self.__fileFullPathComm()
        if os.path.exists(fileNameComm):
            logger.log(
                "WRN", "Code aster comm yet exists !!! Overwrite", self.__logLevel
            )

        try:
            with open(fileNameComm, "w") as f:
                f.write(rendered)  # type: ignore
        except OSError:
            logger.log("ERR", f"Writing file *{fileNameComm}*.", self.__logLevel)
            return False
        else:
            logger.log(
                "INF",
                f"File for aster COMM *{fileNameComm}* created.",
                self.__logLevel,
            )
            self.__fileNameToTrash.append(fileNameComm)

        fileNameRmed = self.__fileFullPathRmed()
        if os.path.exists(fileNameRmed):
            logger.log("WRN", "Code aster yet launched !!! Overwrite", self.__logLevel)

        # Run Code Aster
        #
        rc = self.__launcher.launch(
            fileNameExport, test, fileNameComm, self.__fileFullPathMmed()
        )
        if rc != 0:
            logger.log("ERR", "Code aster failed !!! quit", self.__logLevel)
            return False

        # LOCAL means that we have SINGULARITY container and export will be made
        # by PARAVIS and related plugin for medfile
        #
        if self.__codeAsterLauncher == "LOCAL":

            # Export VTK multiblock file with vtm extension
            #
            if os.path.exists(os.path.join(self.__workingPath, "vtm")):
                logger.log("WRN", "VTK files yet crated !!! Overwrite", self.__logLevel)

            dataArray = ["FamilyIdNode", "resther0TEMP", "FamilyIdCell", "NumIdCell"]
            rc = self.__launcher.exportFromMEDFile(
                fileNameRmed,
                "vtm",
                pointDataArr=dataArray,
                cellDataArr=dataArray,
                timeStep=1,
            )
            if rc != 0:
                logger.log(
                    "ERR", "Export to VTK multiblock failed !!! quit", self.__logLevel
                )
                return False

        # CONTAINER means that we have a container and export will be made
        # by this container and related plugin for medfile
        #
        elif self.__codeAsterLauncher == "CONTAINER":

            # Exporting with option timeStep = -1 means that will be exported
            # complete time series
            rc = self.__launcher.exportFromMEDFile(
                fileNameRmed, pointDataArr=["resther0TEMP"], test=test
            )

            if rc != 0:
                logger.log(
                    "ERR", "Export VTK from MED failed !!! quit", self.__logLevel
                )
                return False

        return True

    def exportThermalImgs(self) -> bool:

        if self.__codeAsterLauncher == "LOCAL":
            # Export images with pvbatch from thermal analysis
            #
            rc = self.__launcher.exportImgsFromVtm(
                self.__workingPath, "vtm", self.__workingPath
            )
            if rc != 0:
                logger.log(
                    "ERR", "Export images from VTK failed !!! quit", self.__logLevel
                )
                return False
            return True
        else:
            logger.log(
                "INF",
                "Exporting thermal images from VTK only in LOCAL mode !!! quit",
                self.__logLevel,
            )
            return True

    def parse(self) -> bool:
        if self.__codeAsterLauncher == "LOCAL":
            if not self.__parser.parse(
                files=os.path.join(self.__workingPath, "vtm"), ll=self.__logLevel
            ):
                logger.log("ERR", "Parsing VTM failed !!! quit", self.__logLevel)
                return False
            else:
                logger.log("INF", "Parsing VTM done.", self.__logLevel)
                return True

        if self.__codeAsterLauncher == "CONTAINER":
            if not self.__parser.parse(
                files=os.path.join(self.__workingPath, "vtu"), ll=self.__logLevel
            ):
                logger.log("ERR", "Parsing VTU failed !!! quit", self.__logLevel)
                return False
            else:
                logger.log("INF", "Parsing VTU done.", self.__logLevel)
                return True

        logger.log("ERR", "Setting launcher wrong !!! quit", self.__logLevel)
        return False

    def getParser(self):
        return self.__parser

    def deleteArtifacts(self):
        # print(self.__fileNameToTrash)
        for f in self.__fileNameToTrash:
            try:
                os.remove(f)
                logger.log("WRN", f"Removed file --> {f}", self.__logLevel)
            except OSError:
                logger.log("ERR", f"Removing file --> {f}", self.__logLevel)
        # print(self.__dirsNameToTrash)
        for d in self.__dirsNameToTrash:
            try:
                shutil.rmtree(d)
                logger.log("WRN", f"Removed dir --> {d}", self.__logLevel)
            except OSError:
                logger.log("ERR", f"Removing dir --> {d}", self.__logLevel)
        self.__fileNameToTrash = []
        self.__dirsNameToTrash = []
        # self.__launcher.deleteArtifacts()

    def buildHotSection(self) -> bool:
        if self.__section is not None:
            # Mesh in [m] but section in [mm]. Need a conversion
            H = self.__section.getDimH() / 1000
            W = self.__section.getDimW() / 1000
        else:
            logger.log("ERR", "Need to use setSection() !!! Quit.", self.__logLevel)
            return False

        if self.__codeAsterLauncher == "LOCAL":

            # Setting actual time value
            #
            logger.log(
                "INF",
                f"Find block with time value {self.__time * 60} ...",
                self.__logLevel,
            )
            k = self.__parser.findMultiblockKeyByArrayValue(
                arrayName="TimeValue", value=self.__time * 60, ll=self.__logLevel
            )
            logger.log("INF", f"... key found {k}.", self.__logLevel)
            self.__parser.setCurrentMultiblockByKey(key=k, ll=self.__logLevel)

        elif self.__codeAsterLauncher == "CONTAINER":

            # Setting index by time step
            #
            timeIndex = int(self.__time / 10)
            logger.log(
                "INF", f"Find block with time index {timeIndex} ...", self.__logLevel
            )
            self.__parser.setCurrentFileByIndex(timeIndex)

        else:
            logger.log(
                "ERR",
                f"Code aster environnment {self.__codeAsterLauncher} wrong",
                self.__logLevel,
            )
            return False

        # Build temperature value in dir. Y
        #
        dirY_y = np.linspace(0, H, self.__meshNumberInY)
        dirY_x = np.zeros(len(dirY_y))
        dirY_v = np.zeros(len(dirY_y))
        for i in range(len(dirY_x)):
            testPoint = Point3d(dirY_x[i], dirY_y[i] - H / 2, 0)
            pointsFound = self.__parser.getPointsCellNearestPoint(
                p=testPoint, ll=self.__logLevel
            )
            dirY_v[i] = self.__parser.getPointValueInterpTriaByIds(
                p=testPoint,
                id0=pointsFound[0].idn,
                id1=pointsFound[1].idn,
                id2=pointsFound[2].idn,
                arrayName="resther0TEMP",
                ll=self.__logLevel,
            )
            logger.log(
                "INF",
                f"---> Dir.Y Temp in point is {dirY_v[i]:.1f}",
                self.__logLevel,
            )

        # Build temperature value in dir. X
        #
        dirX_x = np.linspace(0, W, self.__meshNumberInX)
        dirX_y = np.zeros(len(dirX_x))
        dirX_v = np.zeros(len(dirX_x))
        for i in range(len(dirX_x)):
            testPoint = Point3d(dirX_x[i] - W / 2, dirX_y[i], 0)
            pointsFound = self.__parser.getPointsCellNearestPoint(
                p=testPoint, ll=self.__logLevel
            )
            dirX_v[i] = self.__parser.getPointValueInterpTriaByIds(
                p=testPoint,
                id0=pointsFound[0].idn,
                id1=pointsFound[1].idn,
                id2=pointsFound[2].idn,
                arrayName="resther0TEMP",
                ll=self.__logLevel,
            )
            logger.log(
                "INF",
                f"---> Dir.X Temp in point is {dirX_v[i]:.1f}",
                self.__logLevel,
            )

        zerosDirY, zerosDirY_sgn = self.__findZeros(dirY_y, dirY_v - 500)
        zerosDirX, zerosDirX_sgn = self.__findZeros(dirX_x, dirX_v - 500)

        assert len(zerosDirY) == len(zerosDirY_sgn)

        deltaBot = 0.0
        deltaTop = 0.0
        if len(zerosDirY) in [0, 1, 2]:
            for i in range(len(zerosDirY)):
                if zerosDirY_sgn[i] == -1:
                    deltaBot = zerosDirY[i] * 1000
                elif zerosDirY_sgn[i] == +1:
                    deltaTop = (H - zerosDirY[i]) * 1000
                else:
                    pass
        else:
            logger.log("ERR", "zerosDirY not in [0,1,2] !!! Quit.", self.__logLevel)
            return False

        logger.log(
            "INF",
            f"deltaBot = {deltaBot:.1f} -- deltaTop = {deltaTop:.1f}",
            self.__logLevel,
        )

        deltaLeft = 0.0
        deltaRight = 0.0
        if len(zerosDirX) in [0, 1, 2]:
            for i in range(len(zerosDirX)):
                if zerosDirX_sgn[i] == -1:
                    deltaLeft = zerosDirX[i] * 1000
                elif zerosDirX_sgn[i] == +1:
                    deltaRight = (W - zerosDirX[i]) * 1000
                else:
                    pass
        else:
            logger.log("ERR", "zerosDirX not in [0,1,2] !!! Quit.", self.__logLevel)
            return False

        self.__reductionLeft = deltaLeft
        self.__reductionRight = deltaRight
        self.__reductionTop = deltaTop
        self.__reductionBottom = deltaBot

        logger.log(
            "INF",
            f"deltaLeft = {deltaLeft:.1f} -- deltaRight = {deltaRight:.1f}",
            self.__logLevel,
        )
        logger.log(
            "INF",
            f"deltaTop  = {deltaTop:.1f} --   deltaBot = {deltaBot:.1f}",
            self.__logLevel,
        )

        #
        # Translate concrete part of section after reduction shape
        #
        self.__section.setDimH(self.__section.getDimH() - deltaTop - deltaBot)
        self.__section.setDimW(self.__section.getDimW() - deltaLeft - deltaRight)

        pTL = self.__section.getStructConcretelItem().getShape().getShapePoint("TL")
        pBR = self.__section.getStructConcretelItem().getShape().getShapePoint("BR")

        deltaG_x = ((pBR.x - deltaRight) + (pTL.x + deltaLeft)) / 2
        deltaG_y = ((pTL.y - deltaTop) + (pBR.y + deltaBot)) / 2
        logger.log(
            "INF",
            f"deltaG_x = {deltaG_x:.1f}mm -- deltaG_y = {deltaG_y:.1f}mm",
            self.__logLevel,
        )

        #
        # Assign steel temperature
        #
        if len(self.__section.getSteelRebar()) > 0:
            steelRebarTemperatures = len(self.__section.getSteelRebar()) * [0.0]
            for idx, bar in enumerate(self.__section.getSteelRebar()):
                centerOfBar = bar.getShape().getOrigin()
                #
                # Retrive value in point by interp on cellPoints
                #
                logger.log(
                    "INF",
                    f"Retrive temp by interp on cell points at {centerOfBar}",
                    self.__logLevel,
                )
                testPoint = Point3d(centerOfBar.x / 1000, centerOfBar.y / 1000, 0)

                pointsFound = self.__parser.getPointsCellNearestPoint(
                    p=testPoint, ll=self.__logLevel
                )

                value = self.__parser.getPointValueInterpTriaByIds(
                    p=testPoint,
                    id0=pointsFound[0].idn,
                    id1=pointsFound[1].idn,
                    id2=pointsFound[2].idn,
                    arrayName="resther0TEMP",
                    ll=self.__logLevel,
                )

                if value is not None:
                    logger.log(
                        "INF",
                        f"---> Temp in point is {value:.1f}",
                        self.__logLevel,
                    )
                    steelRebarTemperatures[idx] = value
                else:
                    logger.log(
                        "ERR", "---> Temp in point is None. Quit", self.__logLevel
                    )
                    return False

            self.__section.setSteelTemperatures(steelRebarTemperatures)

            #
            # Translate steel part of section after reduction shape
            # cause center of gravity changed with section reducted
            #
            self.__section.translateSteelItems(Point2d(), Point2d(deltaG_x, deltaG_y))

        else:
            logger.log("INF", "No rebar. Quit.", self.__logLevel)
            return False
        return True

    def __findZeros(self, x: np.ndarray, y: np.ndarray) -> List[List[float]]:
        zeros_val = []
        # Decreasing or increasing funtion
        #
        # decreasing = -1
        # creasing = +1
        # constant = 0
        local_sig = []
        if len(x) != len(y):
            logger.log(
                "ERR",
                "len(x) = {} must be the same of len(y) = {} !!! quit".format(
                    len(x), len(x)
                ),
                self.__logLevel,
            )
            return zeros_val

        for i in range(len(x) - 1):
            if y[i] * y[i + 1] == 0.0:
                if y[i] == 0.0:
                    zeros_val.append(x[i])
                    if y[i + 1] > y[i]:
                        local_sig.append(+1)
                    elif y[i + 1] < y[i]:
                        local_sig.append(-1)
                    else:
                        local_sig.append(0)
            else:
                if y[i] * y[i + 1] < 0.0:
                    zeros_val.append(
                        x[i] + abs(y[i] * (x[i + 1] - x[i]) / (y[i + 1] - y[i]))
                    )
                    if y[i + 1] > y[i]:
                        local_sig.append(+1)
                    elif y[i + 1] < y[i]:
                        local_sig.append(-1)
                    else:
                        local_sig.append(0)

        return [zeros_val, local_sig]

    def getDeltas(self) -> List[float]:
        return [
            self.__reductionLeft,
            self.__reductionRight,
            self.__reductionTop,
            self.__reductionBottom,
        ]



"""
Created on Wed Jan 16 18:02:25 2019

@author: lpaone
"""
from __future__ import annotations

import math
from copy import deepcopy
from typing import List, Optional, Tuple, Union

from pydantic import BaseModel


# Defines points or free vector in 2D space
class Point2d:
    def __init__(
        self,
        x: Union[float, int] = 0.0,
        y: Union[float, int] = 0.0,
        make_null: bool = False,
        coords: Union[Tuple[Union[float, int], Union[float, int]], None] = None,
    ):
        if isinstance(x, (float, int)):
            self.__x = x
            self.__y = y
            self.__isNull: bool = make_null
        else:
            raise ValueError("Point2d init only with float or int !!!")

        if coords is not None:
            self.__x = coords[0]
            self.__y = coords[1]

    def __str__(self):
        return f"Point2d: ({self.__x:.2e},{self.__y:.2e})"

    def __add__(self, other: Point2d) -> Point2d:
        return Point2d(self.__x + other.x, self.__y + other.y)

    def __sub__(self, other: Point2d) -> Point2d:
        return Point2d(self.__x - other.x, self.__y - other.y)

    def __mul__(self, other: Union[float, int]) -> Point2d:
        return Point2d(self.__x * other, self.__y * other)

    def __rmul__(self, other: float) -> Point2d:
        return Point2d(self.__x * other, self.__y * other)

    def __eq__(self, other: Point2d):
        if self.__x == other.x and self.__y == other.y:
            return True
        else:
            return False

    def __repr__(self):
        return str(self)

    @property
    def x(self):
        return self.__x

    @x.setter
    def x(self, val: float) -> None:
        self.__x = val

    @property
    def y(self):
        return self.__y

    @y.setter
    def y(self, val: float) -> None:
        self.__y = val

    def translate(self, dx=0.0, dy=0.0):

        assert type(self.__x) is float or type(self.__x) is int
        assert type(self.__y) is float or type(self.__y) is int

        self.__x = self.__x + dx
        self.__y = self.__y + dy

    def isNull(self) -> bool:
        return self.__isNull

    def isEqualTo(self, other, rox, roy, prec):
        dx = math.sqrt(math.pow(self.__x - other.x, 2) / math.pow(rox, 2))
        dy = math.sqrt(math.pow(self.__y - other.y, 2) / math.pow(roy, 2))
        if dx < prec and dy < prec:
            return True
        else:
            return False

    def distance(self, other):
        return math.sqrt(
            math.pow(self.__x - other.x, 2) + math.pow(self.__y - other.y, 2)
        )

    def midpoint(self, other):
        return Point2d((self.__x + other.x) / 2, (self.__y + other.y) / 2)

    def cross(self, p1, p2):
        x1 = p1.x - self.__x
        y1 = p1.y - self.__y

        x2 = p2.x - self.__x
        y2 = p2.y - self.__y

        return x1 * y2 - x2 * y1

    @staticmethod
    def areaFromTria(p0: Point2d, p1: Point2d, p2: Point2d) -> float:
        assert isinstance(p0.x, (float, int))
        assert isinstance(p0.y, (float, int))
        assert isinstance(p1.x, (float, int))
        assert isinstance(p1.y, (float, int))
        assert isinstance(p2.x, (float, int))
        assert isinstance(p2.y, (float, int))
        return (
            1
            / 2
            * abs(
                p0.x * p1.y * 1
                + p0.y * 1 * p2.x
                + 1 * p1.x * p2.y
                - 1 * p1.y * p2.x
                - p0.y * p1.x * 1
                - p0.x * 1 * p2.y
            )
        )


def boundingBox(p1, p2):
    if not isinstance(p1, Point2d) or not isinstance(p2, Point2d):
        raise Exception("Bounding only for two Point2d !!!")

    assert isinstance(p1.x, (float, int))
    assert isinstance(p1.y, (float, int))
    assert isinstance(p2.x, (float, int))
    assert isinstance(p2.y, (float, int))

    min_x = min(p1.x, p2.x)
    max_x = max(p1.x, p2.x)
    min_y = min(p1.y, p2.y)
    max_y = max(p1.y, p2.y)

    return [min_x, max_x, min_y, max_y]


class Node2d(Point2d):
    def __init__(self, x=0.0, y=0.0, idn=-1):
        Point2d.__init__(self, x, y)
        self.idn = idn

    def __str__(self):
        return f"Node2D: idn = {self.idn:.0f}, x = {self.__x:.3e}, y = {self.__y:.3e}"


# Defines points or free vector in 3D space
class Point3d:
    def __init__(self, x=0.0, y=0.0, z=0.0):
        self.__x = x
        self.__y = y
        self.__z = z

    def __str__(self):
        return f"Point3D: x = {self.__x:.3e}, y = {self.__y:.3e}, z = {self.__z:.3e}"

    def __add__(self, other) -> Point3d:
        return Point3d(self.__x + other.x, self.__y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Point3d(self.__x - other.x, self.__y - other.y, self.z - other.z)

    @property
    def x(self):
        return self.__x

    @x.setter
    def x(self, val: float) -> None:
        self.__x = val

    @property
    def y(self):
        return self.__y

    @y.setter
    def y(self, val: float) -> None:
        self.__y = val

    @property
    def z(self):
        return self.__z

    @z.setter
    def z(self, val: float) -> None:
        self.__z = val

    def tranlate(self, dx=0.0, dy=0.0, dz=0.0):
        self.__x = self.__x + dx
        self.__y = self.__y + dy
        self.z = self.z + dz

    def distanceFrom(self, p: Point3d) -> float:

        if not isinstance(p, Point3d):
            raise Exception("Arg must be a Point3d instance !!!")

        return math.dist([self.__x, self.__y, self.z], [p.x, p.y, p.z])

    # @staticmethod
    # def areaFromTria(p0, p1, p2) -> float:
    # area0 = Point2d.areaFromTria(Point2d(p0.x,p0.y),Point2d(p1.x,p1.y),Point2d(p2.x,p2.y))
    # area1 = Point2d.areaFromTria(Point2d(p0.y,p0.z),Point2d(p1.y,p1.z),Point2d(p2.y,p2.z))
    # area2 = Point2d.areaFromTria(Point2d(p0.x,p0.z),Point2d(p1.x,p1.z),Point2d(p2.x,p2.z))
    # return 1/2 * math.sqrt(area0*area0 + area1*area1 + area2*area2)


def areaFromTria3D(p0: Point3d, p1: Point3d, p2: Point3d) -> float:
    v = Vector3d(p0, p1)
    w = Vector3d(p1, p2)
    return 1 / 2 * math.sqrt(abs(v.scalar(v) * w.scalar(w) - v.scalar(w) * v.scalar(w)))


class Node3d(Point3d):
    def __init__(self, x=0.0, y=0.0, z=0.0, idn=-1):
        Point3d.__init__(self, x, y, z)
        self.idn = idn

    def __str__(self):
        return "Node3D: idn = {:.0f}, x = {:.3e}, y = {:.3e}, z = {:.3e}".format(
            self.idn,
            self.x,
            self.y,
            self.z,
        )

    def __add__(self, other) -> Node3d:
        return Node3d(self.__x + other.x, self.__y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Node3d(self.__x - other.x, self.__y - other.y, self.z - other.z)


class Seg2d:
    def __init__(self, p0: Point2d, p1: Point2d):
        self.__p0 = p0
        self.__p1 = p1

    @property
    def p0(self):
        return self.__p0

    @p0.setter
    def p0(self, val: Point2d) -> None:
        self.__p0 = val

    @property
    def p1(self):
        return self.__p1

    @p1.setter
    def p1(self, val: Point2d) -> None:
        self.__p1 = val

    def len(self) -> float:
        return self.__p0.distance(self.__p1)

    # Find the intersection point by Seg2D in input
    def intersect(self, other: Seg2d) -> Point2d:
        a0 = self.__p1.x - self.__p0.x
        b0 = other.p0.x - other.p1.x
        c0 = other.p0.x - self.__p0.x

        a1 = self.__p1.y - self.__p0.y
        b1 = other.p0.y - other.p1.y
        c1 = other.p0.y - self.__p0.y

        detA = a0 * b1 - a1 * b0
        if detA == 0:
            return Point2d(make_null=True)

        detA_alpha = c0 * b1 - c1 * b0
        alpha = detA_alpha / detA
        return self.__p0 + alpha * (self.__p1 - self.__p0)

    def boundingHasPoint(self, p: Point2d, toll: float = 1e-15):
        bb = boundingBox(self.__p0, self.__p1)
        if (
            p.x < bb[0] - toll
            or p.x > bb[1] + toll
            or p.y < bb[2] - toll
            or p.y > bb[3] + toll
        ):
            return False
        else:
            return True

    # Calculate intersection point from extension of other and self
    def intersectAndContaints(self, other: Seg2d) -> Tuple[bool, Point2d]:
        pint = self.intersect(other)
        m1 = (pint.distance(self.__p0) + pint.distance(self.__p1))/ self.__p0.distance(self.__p1)
        # m2 = (pint.distance(other.p0) + pint.distance(other.p1)) / other.len()
        pint.isNull()
        if abs(m1-1) < 0.0000000001:
            return True, pint
        else:
            return False, pint

        # if not pint.isNull() and self.boundingHasPoint(pint, toll=self.len() / 100000):
        #     return True, pint
        # else:
        #     return False, pint

    # If p is a Point2d calculs factor witch factor * Seg2d = p
    def extensionFactor(self, p):
        vx = self.__p0.x - self.__p1.x
        vy = self.__p0.y - self.__p1.y

        evx = p.x - self.__p1.x
        evy = p.y - self.__p1.y

        f = 1
        if vx != 0.0:
            f = evx / vx
        if vy != 0.0:
            f = evy / vy
        return f

    def __str__(self):
        dispstr = "Point2d i: x = {:.3e}, y = {:.3e} \n".format(
            self.__p0.x,
            self.__p0.y,
        )
        dispstr = dispstr + "Point2d j: x = {:.3e}, y = {:.3e} \n".format(
            self.__p1.x,
            self.__p1.y,
        )
        return dispstr


class Edge2d:
    def __init__(self, *args):
        if len(args) == 2:
            if isinstance(args[0], Node2d) & isinstance(args[1], Node2d) is True:
                self.__node_i = args[0]
                self.__node_j = args[1]
            else:
                raise TypeError("Args must be [Node2D] !!!")
        else:
            raise TypeError("Only two argument !!!")

    def nodeI(self) -> Node2d:
        return self.__node_i

    def nodeJ(self) -> Node2d:
        return self.__node_j

    def __str__(self):
        dispstr = "Node2D i: idn = {:.0f}, x = {:.3e}, y = {:.3e} \n".format(
            self.__node_i.idn,
            self.__node_i.x,
            self.__node_i.y,
        )
        dispstr = dispstr + "Node2D j: idn = {:.0f}, x = {:.3e}, y = {:.3e} \n".format(
            self.__node_j.idn,
            self.__node_j.x,
            self.__node_j.y,
        )
        return dispstr


class Edge3d:
    def __init__(self, *args):
        if len(args) == 2:
            if isinstance(args[0], Node3d) & isinstance(args[1], Node3d) is True:
                self.__node_i = args[0]
                self.__node_j = args[1]
            else:
                raise Exception("Args must be [Node3D] !!!")
        elif len(args) == 0:
            self.__node_i = Node3d()
            self.__node_j = Node3d()
        elif len(args) == 8:
            for i in range(2):
                if isinstance(args[0 + 4 * i], float) is False:
                    raise Exception("Arg n. %e must be float !!!" % (0 + i))
                if isinstance(args[1 + 4 * i], float) is False:
                    raise Exception("Arg n. %e must be float !!!" % (1 + i))
                if isinstance(args[2 + 4 * i], float) is False:
                    raise Exception("Arg n. %e must be float !!!" % (2 + i))
                if isinstance(args[3 + 4 * i], int) is False:
                    raise Exception("Arg n. %e must be int !!!" % (3 + i))
            self.__node_i = Node3d(args[0], args[1], args[2], args[3])
            self.__node_j = Node3d(args[4], args[5], args[6], args[7])
        else:
            raise Exception("Wrong arguments !!!")

    def nodeI(self) -> Node3d:
        return self.__node_i

    def nodeJ(self) -> Node3d:
        return self.__node_j

    def setNodeI(self, *args):
        if len(args) == 1:
            if isinstance(args[0], Node3d) is True:
                self.__node_i = args[0]
            else:
                raise Exception("Arg must be [Node3D] !!!")
        else:
            raise Exception("Only one argument !!!")

    def setNodeJ(self, *args):
        if len(args) == 1:
            if isinstance(args[0], Node3d) is True:
                self.__node_j = args[0]
            else:
                raise Exception("Arg must be [Node3D] !!!")
        else:
            raise Exception("Only one argument !!!")

    def __str__(self):
        dispstr = (
            "Node3D i: idn = {:.0f}, x = {:.3e}, y = {:.3e}, z = {:.3e} \n".format(
                self.__node_i.idn,
                self.__node_i.x,
                self.__node_i.y,
                self.__node_i.z,
            )
        )
        dispstr = (
            dispstr
            + "Node3D j: idn = {:.0f}, x = {:.3e}, y = {:.3e}, z = {:.3e} \n".format(
                self.__node_j.idn,
                self.__node_j.x,
                self.__node_j.y,
                self.__node_j.z,
            )
        )
        return dispstr

    def lenght(self):
        return math.sqrt(
            math.pow(self.__node_j.x - self.__node_i.x, 2)
            + math.pow(self.__node_j.y - self.__node_i.y, 2)
            + math.pow(self.__node_j.z - self.__node_i.z, 2)
        )


class Polyline2d:
    """
    List of Nodes2D
    """

    def __init__(self, *args):
        if len(args) == 0:
            raise TypeError("Only one argument !!!")
        elif len(args) == 1:
            lst = args[0]
            if isinstance(lst, list) is True:
                for i in lst:
                    if isinstance(i, Node2d) is False:
                        raise TypeError("List must be [Node2D] formed !!!")
                self.__node = lst
            else:
                raise TypeError("First arg must be type List !!!")
        elif len(args) > 1:
            raise TypeError("Only one argument !!!")

    def __str__(self):
        dispstr = "Polyline2d: size is %.0f \n" % len(self.__node)
        for n in self.__node:
            dispstr = (
                dispstr
                + "Node2D: idn = {:.0f}, x = {:.3e}, y = {:.3e} \n".format(
                    n.idn,
                    n.x,
                    n.y,
                )
            )
        return dispstr

    def getNodes(self):
        return self.__node

    def size(self) -> int:
        return len(self.__node)

    def setClosed(self):
        self.__node.append(self.__node[0])
        return

    def isClosed(self):
        sz = len(self.__node)
        if self.__node[0] == self.__node[sz - 1]:
            return True
        else:
            return False

    def translate(
        self, pStart: Union[Point2d, None], pEnd: Union[Point2d, None]
    ) -> None:
        if pStart is None:
            pStart = Point2d()
        if pEnd is None:
            pEnd = Point2d()

        v = pEnd - pStart
        for node in self.__node:
            node.translate(v.x, v.y)

    def vertexNb(self):
        return len(self.__node)

    def vertexAt(self, i):
        return len(self.__node[i])


class Polyline3d:
    """
    List of Nodes3D
    """

    def __init__(self, *args):
        if len(args) == 0:
            raise Exception("Only one argument !!!")
        elif len(args) == 1:
            lst = args[0]
            if isinstance(lst, list) is True:
                for i in lst:
                    if isinstance(i, Node3d) is False:
                        raise Exception("List must be [Node3D] formed !!!")
                self.__node = lst
            else:
                raise Exception("First arg must be type List !!!")
        elif len(args) > 1:
            raise Exception("Only one argument !!!")

    def __str__(self):
        dispstr = "Polyline3d: size is %.0f \n" % len(self.__node)
        for n in self.__node:
            dispstr = (
                dispstr
                + "Node3D: idn = {:.0f}, x = {:.3e}, y = {:.3e}, z = {:.3e}\n".format(
                    n.idn,
                    n.x,
                    n.y,
                    n.z,
                )
            )
        return dispstr

    def setClosed(self):
        self.__node.append(self.__node[0])
        return

    def isClosed(self):
        sz = len(self.__node)
        if self.__node[0] == self.__node[sz - 1]:
            return True
        else:
            return False

    def vertexNb(self):
        return len(self.__node)

    def vertexAt(self, i):
        return len(self.__node[i])


# Defines points or free vector in 3D space
class Vector2d:
    def __init__(
        self,
        p0: Point2d = Point2d(make_null=True),
        p1: Point2d = Point2d(make_null=True),
        vx: float = 0.0,
        vy: float = 0.0,
    ):

        if isinstance(p0, Point2d) and isinstance(p1, Point2d):
            if not p0.isNull() and not p1.isNull():
                self.vx = p1.x - p0.x
                self.vy = p1.y - p0.y
                return
        else:
            raise TypeError("p0 and p1 must be Point2d instances")

        self.vx = vx
        self.vy = vy

    def __str__(self):
        return f"Vector2D: vx = {self.vx:.3e}, vy = {self.vy:.3e}"

    def __add__(self, other: Vector2d) -> Vector2d:
        return Vector2d(vx=self.vx + other.vx, vy=self.vy + other.vy)

    def __sub__(self, other: Vector2d):
        return Vector2d(vx=self.vx - other.vx, vy=self.vy - other.vy)

    # def __mul__(self, scale):
    #    return Vector3d(self.vx*scale,self.vy*scale,self.z*scale)

    def __rmul__(self, scale):
        return Vector2d(vx=self.vx * scale, vy=self.vy * scale)

    #       _    _    _            _         _         _
    #     | x    y    z  |   y1*z2*x + z1*x2*y + x1*y2*z +
    #  det| x1   y1   z1 | =       _         _         _
    #     | x2   y2   z2 |  -y2*z1*x - x1*z2*y - x2*y1*z
    #
    def cross(self, other: Vector2d):
        x1 = self.vx
        y1 = self.vy

        x2 = other.vx
        y2 = other.vy

        return x1 * y2 - x2 * y1

    def norm(self):
        return math.sqrt(math.pow(self.vx, 2) + math.pow(self.vy, 2))

    def norm_roxroy(self, rox=1.0, roy=1.0):
        return math.sqrt(math.pow(self.vx / rox, 2) + math.pow(self.vy / roy, 2))

    def normalize(self):
        n = self.norm()
        if n != 0:
            self.vx = self.vx / n
            self.vy = self.vy / n
        else:
            raise Exception("Vector 2D null !!!")
        return self

    def rotate(self, angle: float) -> Vector2d:
        hold_vx = self.vx
        hold_vy = self.vy
        alpha = math.radians(angle)
        self.vx = (math.cos(alpha)) * hold_vx + (-math.sin(alpha)) * hold_vy
        self.vy = (math.cos(alpha)) * hold_vy + (math.sin(alpha)) * hold_vx

        return self


# Defines points or free vector in 3D space
class Vector3d:
    def __init__(self, *args):
        if len(args) == 3:
            for a in args:
                if isinstance(a, float) is False and isinstance(a, int) is False:
                    raise Exception("With len args == 3 only float or int type !!!")

            self.vx = float(args[0])
            self.vy = float(args[1])
            self.vz = float(args[2])

        elif len(args) == 2:
            for a in args:
                if isinstance(a, Point3d) is False:
                    raise Exception("With len args == 2 only Point3d type !!!")

            self.vx = args[1].x - args[0].x
            self.vy = args[1].y - args[0].y
            self.vz = args[1].z - args[0].z

        elif len(args) == 0:
            self.vx = 0.0
            self.vy = 0.0
            self.vz = 0.0

        else:
            raise Exception("Wrong args !!!")

    def __str__(self):
        return f"Vector3D: vx = {self.vx:.3e}, vy = {self.vy:.3e}, vz = {self.vz:.3e}"

    def __add__(self, other: Vector3d) -> Vector3d:
        return Vector3d(self.vx + other.vx, self.vy + other.vy, self.vz + other.vz)

    def __sub__(self, other: Vector3d) -> Vector3d:
        return Vector3d(self.vx - other.vx, self.vy - other.vy, self.vz - other.vz)

    # def __mul__(self, scale):
    #    return Vector3d(self.vx*scale,self.vy*scale,self.z*scale)

    def __rmul__(self, scale):
        return Vector3d(self.vx * scale, self.vy * scale, self.vz * scale)

    #       _    _    _            _         _         _
    #     | x    y    z  |   y1*z2*x + z1*x2*y + x1*y2*z +
    #  det| x1   y1   z1 | =       _         _         _
    #     | x2   y2   z2 |  -y2*z1*x - x1*z2*y - x2*y1*z
    #
    def cross(self, other):
        x1 = self.vx
        y1 = self.vy
        z1 = self.vz
        x2 = other.vx
        y2 = other.vy
        z2 = other.vz
        x3 = y1 * z2 - y2 * z1
        y3 = z1 * x2 - x1 * z2
        z3 = x1 * y2 - x2 * y1
        return Vector3d(x3, y3, z3)

    def norm(self):
        return math.sqrt(
            math.pow(self.vx, 2) + math.pow(self.vy, 2) + math.pow(self.vz, 2)
        )

    def normalize(self):
        n = self.norm()
        if n != 0:
            self.vx = self.vx / n
            self.vy = self.vy / n
            self.vz = self.vz / n
        else:
            raise Exception("Vector 3D null !!!")
        return self

    def scalar(self, other):
        return self.vx * other.vx + self.vy * other.vy + self.vz * other.vz


# Sum vector to point in affine space
def affineSum2d(p: Point2d, v: Vector2d) -> Point2d:
    return Point2d(p.x + v.vx, p.y + v.vy)


# Make a points array with equal space between theme
# Extremes stand for include first and last point
def twoPointsDivide(
    p0: Point2d, p1: Point2d, nb: int, extremes: bool = True
) -> List[Point2d]:
    if nb < 1:
        raise ValueError("nb must be >=1")
    lPoints = []
    if extremes:
        lPoints.append(deepcopy(p0))
    vDir = Vector2d(p0, p1).normalize()
    space = p0.distance(p1) / nb
    for i in range(1, nb):
        lPoints.append(affineSum2d(p0, space * i * vDir))
    if extremes:
        lPoints.append(deepcopy(p1))
    return lPoints


def twoPointsMiddle(p0: Point2d, p1: Point2d) -> Point2d:
    return Point2d((p0.x + p1.x) / 2, (p0.y + p1.y) / 2)


# Copy of points translated with offset
def twoPointsOffset(
    p0: Point2d, p1: Point2d, offset: float = 0.0
) -> Tuple[Point2d, Point2d]:
    nDir = offset * Vector2d(p0, p1).normalize().rotate(90)
    return affineSum2d(p0, nDir), affineSum2d(p1, nDir)


class Point2dMdl(BaseModel):
    xy: Optional[Tuple[float, float]] = None



import warnings
from enum import Enum
from typing import Any, Dict, List, Union

from pycivil.EXAUtils.EXAExceptions import EXAExceptions as Ex


class Table:
    def __init__(self, data: Dict[str, Any]) -> None:
        self.__data = data
        self.__choose = -1

        # for iterables
        self.__i = -1

    def __len__(self):
        return len(self.__data["rows"])

    def __iter__(self):
        return self

    def __next__(self):
        if self.__i < self.__len__() - 1:
            self.__i += 1
            return self.__data["rows"][self.__i]
        else:
            raise StopIteration()

    def setChoose(self, idr: int) -> None:

        if "columns" in self.__data.keys():
            if "id" in self.__data["columns"]:
                idx = self.__data["columns"].index("id")
            else:
                raise Ex("0001", "one column must have id key")
        else:
            raise Ex("0002", "data must have columns key")

        for r in self.__data["rows"]:
            if r[idx] == idr:
                self.__choose = idr
                print("Scelta -->", r[1])
                return

        raise Ex("0002", "not found. wrong choose")

    def choose(self):
        return self.__choose

    def data(self):
        return self.__data

    def valueAt(self, idx: Any = -1) -> Union[Dict[str, Any], List[Any]]:

        if not isinstance(idx, int) and not isinstance(idx, Enum):
            raise Ex("ERR", "id must be int or Enum", type(idx))

        if idx is Enum:
            idx = idx.value  # type: ignore

        if idx == -1 and self.__choose == -1:
            print("ERR: not found. Use setChoose(self,id) with correct id !!!")
            return {}

        if idx == -1:
            idx = self.__choose

        for r in self.__data["rows"]:
            if r[0] == idx:
                return r

        raise Ex("ERR", "not found in tables index", idx)

    def tableValue(self):
        warnings.warn("tableValue() will be removed. Use value() instead !!!")
        if self.__choose == -1:
            print("ERR: Use setChoose(self,id) with correct id !!!")
            return ()
        else:
            for r in self.__data["rows"]:
                if r[0] == self.__choose:
                    return r

    def value(self):
        if self.__choose == -1:
            print("ERR: Use setChoose(self,id) with correct id !!!")
            return ()
        else:
            for r in self.__data["rows"]:
                if r[0] == self.__choose:
                    return r

    def __str__(self):
        dispstr = "Table Object: \n"
        dispstr = dispstr + "--------------- \n"
        if "source" in self.__data.keys():
            dispstr = dispstr + " source: {source}\n".format(
                source=self.__data["source"]
            )
        else:
            dispstr = dispstr + " source: ..."

        if "columns" in self.__data.keys():
            dispstr = dispstr + "columns: {columns}\n".format(
                columns=self.__data["columns"]
            )
        else:
            dispstr = dispstr + "columns: ..."

        if "udm" in self.__data.keys():
            dispstr = dispstr + "    udm: {udm}".format(udm=self.__data["udm"])
        else:
            dispstr = dispstr + "    udm: ..."

        dispstr += "\n"
        if "rows" in self.__data.keys():
            for row in self.__data["rows"]:
                if self.__choose != row[0]:
                    dispstr = dispstr + f"    row: {row}\n"
                else:
                    dispstr = dispstr + f"-->  row: {row}\n"
        else:
            dispstr = dispstr + "    row: ..."

        return dispstr



import unittest

from pycivil.EXAGeometry.shapes import Point2d, ShapeArea, ShapeCircle, ShapeRect
from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.materials import Concrete, ConcreteSteel
from pycivil.EXAStructural.sections import ConcreteSection, StructSectionItem


class Test(unittest.TestCase):
    def test_001_rc_rectangular(self):
        bar = ShapeCircle(300)
        print(bar)

        area_shape = ShapeArea(1000)
        print(area_shape)

        # ******************* CONCRETE SECTION first method

        # Setting working code
        code_EC2 = Code("EC2")
        print(code_EC2)

        # Setting concrete material
        cls_material = Concrete("EC2_C20/25")
        cls_material.setByCode(code_EC2, "C20/25")
        print(cls_material)

        # Setting concrete shape
        rect_shape = ShapeRect(300, 500)
        print(rect_shape)

        # Test iterate over vertex coords
        print("Max in x is %1.f" % rect_shape.vertexMaxInX()[0])
        print("Max in y is %1.f" % rect_shape.vertexMaxInY()[0])
        print("Min in x is %1.f" % rect_shape.vertexMinInX()[0])
        print("Min in y is %1.f" % rect_shape.vertexMinInY()[0])

        rectangularSection = StructSectionItem(rect_shape, cls_material)
        print(rectangularSection)

        # Setting steel material
        steel_material = ConcreteSteel("EC2_450C")
        steel_material.setByCode(code_EC2, "B450C")
        print(steel_material)

        # Testing special vertex
        TL = rectangularSection.getShape().getShapePoint("TL")
        print(TL)
        TR = rectangularSection.getShape().getShapePoint("TR")
        print(TR)
        BL = rectangularSection.getShape().getShapePoint("BL")
        print(BL)
        BR = rectangularSection.getShape().getShapePoint("BR")
        print(BR)
        MB = rectangularSection.getShape().getShapePoint("MB")
        print(MB)
        MT = rectangularSection.getShape().getShapePoint("MT")
        print(MT)
        G = rectangularSection.getShape().getShapePoint("G")
        print(G)

        # Setting steel shape
        area_shape_1 = ShapeArea(200)
        area_shape_1.setOrigin(MB + Point2d(0, 50))

        steel_1 = StructSectionItem(area_shape_1, steel_material)
        print(steel_1)

        area_shape_2 = ShapeArea(200)
        area_shape_2.setOrigin(MT + Point2d(0, -50))

        steel_2 = StructSectionItem(area_shape_2, steel_material)
        print(steel_2)

        # TODO: __str__ for ConcreteSection
        myfirstsection = ConcreteSection(1, "300x500 EC2")
        myfirstsection.setStructConcrItem(rectangularSection)
        myfirstsection.setStructSteelItems([steel_1, steel_2])

        print(myfirstsection)

        print("Min steel items !!!")
        print(myfirstsection.findLowSteelItem())
        print("Max steel items !!!")
        print(myfirstsection.findHitSteelItem())
        print("Steel top recover !!!")
        print(myfirstsection.getSteelTopRecover())
        print("Steel bot recover !!!")
        print(myfirstsection.getSteelTopRecover())


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from typing import List

from pydantic import BaseModel


class CodeEnum(str, Enum):
    NTC2008 = "NTC2008"
    NTC2018 = "NTC2018"
    EC2 = "EC2"
    EC2ITA = "EC2:ITA"
    NTC2018RFI = "NTC2018:RFI"


class CodeInfo(BaseModel):
    """Code information."""

    code: str
    description: str
    author: str
    year: int
    country: str

    def __str__(self):
        """Verbose representation of the code."""
        return (
            f"Code Object: \n"
            f"------------ \n"
            f"    id: {str(self.code)}\n"
            f" descr: {self.description}\n"
        )

    @classmethod
    def from_code(cls, code: str) -> "CodeInfo":
        """
        Returns the code info for the given code

        Args:
            code: Code key

        Returns:
            Code information

        Raises:
            ValueError: on unknown code
        """
        try:
            return KNOWN_CODES[code]
        except KeyError as err:
            raise ValueError(f"Unknown code: {code}") from err


KNOWN_CODES: dict[str, CodeInfo] = {
    "EC2": CodeInfo(
        code="EC2",
        description="Eurocode 2: Design of concrete structures - 2005 Edition",
        author="C.E.N.",
        year=2005,
        country="ITA",
    ),
    "EC2:ITA": CodeInfo(
        code="EC2:ITA",
        description="Eurocodice 2: Progettazione delle strutture in calcestruzzo - 2005 Edition",
        author="U.N.I.",
        year=2005,
        country="ITA",
    ),
    "EC2-1-1:2005:ITA": CodeInfo(
        code="EC2-1-1:2005:ITA",
        description="Eurocode 2 Parte 1-1: Progettazione delle strutture in calcestruzzo – Regole generali e regole "
        "per gli edifici – Edizione del 2005",
        author="U.N.I.",
        year=2005,
        country="ITA",
    ),
    "NTC2008": CodeInfo(
        code="NTC2008",
        description="Norme tecniche per le costruzioni di cui al DM 14 gennaio 2008",
        author="M.I.T.",
        year=2008,
        country="ITA",
    ),
    "NTC2018": CodeInfo(
        code="NTC2018",
        description="Norme tecniche per le costruzioni di cui al DM 17 gennaio 2018",
        author="M.I.T.",
        year=2018,
        country="ITA",
    ),
    "CIRC2008": CodeInfo(
        code="CIRC2008",
        description='Circolare 2 febbraio 2009, n. 617 - Istruzioni per l\'applicazione delle "Nuove norme tecniche per le costruzioni" di cui al D.M. 14 gennaio 2008',
        author="C.S.LL.PP.",
        year=2009,
        country="ITA",
    ),
    "CIRC2018": CodeInfo(
        code="CIRC2018",
        description='Circolare 21 gennaio 2019, n. 2 - Istruzioni per l\'applicazione dell\' "Aggiornamento delle Norme tecniche per le costruzioni" di cui al D.M. 17 gennaio 2018',
        author="C.S.LL.PP.",
        year=2019,
        country="ITA",
    ),
}


class Code:
    """Class to select technical standards

    Class to build code. The code is a selector for technical standars in
    construction. E.g. EC2, NTC2008, NTC2018

    Args:
        codeStr (str): Key to select code

    """

    tabKeys = (
        "EC2",
        "EC2:ITA",
        "NTC2008",
        "NTC2018",
        "CIRC2008",
        "CIRC2018",
        "NTC2018:RFI",
    )
    tabIndices = (0, 1, 2, 3, 4, 5, 6)
    tabDescriptions = (
        "Eurocode 2: Design of concrete structures - 2005 Edition",
        "Eurocode 2: modificato con DAN 20 luglio 2007",
        "Norme tecniche per le costruzioni di cui al DM 14 gennaio 2008",
        "Norme tecniche per le costruzioni di cui al DM 17 gennaio 2018",
        "Circolare esplicativa di cui al DM 14 gennaio 2008",
        "Circolare esplicativa di cui al DM 17 gennaio 2018",
        "Norme tecniche per le costruzioni di cui al DM 17 gennaio 2018 modificate da RFI",
    )

    def __init__(self, codeStr: str = "EC2"):
        self.__i = -1
        if isinstance(codeStr, str):
            if codeStr in self.tabKeys:
                self.__byCode = codeStr
            else:
                raise ValueError(f"Code string {codeStr} not in tabKeys !!!")
        else:
            raise ValueError("Only one str argument !!!")

    def __len__(self):
        return len(self.tabKeys)

    def __iter__(self):
        return self

    def __next__(self):
        if self.__i < self.__len__() - 1:
            self.__i += 1
            return self.tabKeys[self.__i]
        else:
            raise StopIteration()

    def setCodeStr(self, codeStr):
        if isinstance(codeStr, str):
            if codeStr in self.tabKeys:
                self.__byCode = codeStr
            else:
                print("WRN: str not in list !!!")
                self.__byCode = ""
        else:
            raise Exception("Only one str argument !!!")

    def codeStr(self):
        return self.__byCode

    def codeDescr(self, i=-1):
        if i == -1:
            if self.__byCode != "":
                return self.tabDescriptions[self.tabKeys.index(self.__byCode)]
            else:
                return ""
        if i >= 0 and i < len(self.tabKeys):
            return self.tabDescriptions[i]
        else:
            return ""

    def __str__(self):
        dispstr = "Code Object: \n"
        dispstr = dispstr + "--------------- \n"
        dispstr = dispstr + "    id: " + str(self.__byCode) + "\n"
        if self.__byCode in self.tabKeys:
            dispstr = (
                dispstr
                + " descr: "
                + str(self.tabDescriptions[self.tabKeys.index(self.__byCode)])
                + "\n"
            )
        else:
            dispstr = dispstr + " descr: " + "\n"
        return dispstr


class Codes:
    """
    Class to build list of Code.
    """

    def __init__(self, codes: List[Code] = []) -> None:
        if codes is None:
            codes = []
        self.__codes = codes

    def __str__(self):
        dispstr = ""
        for c in self.__codes:
            dispstr += c.__str__()

        return dispstr

    def appendUniqueCode(self, code: Code | None = None, code_str = "") -> bool:
        if code is not None:
            for c in self.__codes:
                if c.codeStr() == code.codeStr():
                    return False
            self.__codes.append(code)
            return True
        if code_str != "":
            for c in self.__codes:
                if c.codeStr() == code_str:
                    return False
            self.__codes.append(Code(code_str))
            return True
        return False

    def getCodes(self) -> List[Code]:
        return self.__codes

    def len(self) -> int:
        return len(self.__codes)



# Drawing interation domain and evidenziate intersection on plot diagram

from pycivil.EXAStructural import templateRCRect as est

# Setting new instance of section with
# id = 1 and name = " testing intersection"
section = est.RCTemplRectEC2(1, "First Section testing intersection")

section.setDimH(500.0)
section.setDimW(300.0)

# 'MB' means medium bottom area
# 'MT' means medium top area
section.addSteelArea("MB", 40.0, 1000.0)
section.addSteelArea("MT", 40.0, 0.0)

section.setMaterials("C32/40", "B450C")

section.getMaterialConcr().set_alphacc(1.00)

# Adding Tension Point

pt = section.addTensionPoint2d(N=5.000e05, M=2.103e08)
p0 = section.addTensionPoint2d(N=000.0 * 1e3, M=000.0 * 1e6)

center = p0

# Building a set points of interaction diagramm
pointsCloud, bounding = section.interactionDomainBuild2d(nbPoints=200, bounding=True)

brox = bounding[1] - bounding[0]
broy = bounding[3] - bounding[2]

contained, pintersect, intfactor, other = pointsCloud.contains(
    pt.x, pt.y, rayFromCenter=True, center=p0, ro=(brox, broy)
)

linesOfMinimalPoints = []

if len(pintersect) == 2:
    section.addTensionPoint2d(pintersect[0])
    section.addTensionPoint2d(pintersect[1])
    linesOfMinimalPoints.append([pt, pintersect[0]])
    linesOfMinimalPoints.append([pt, pintersect[1]])
    print("intfactor = %1.6f" % intfactor)
    print("Punti di intersezione ....................")
    print(pintersect[0])
    print(pintersect[1])

section.addTensionPoint2d(center)

section.interactionDomainPlot2d(lines=linesOfMinimalPoints)

print("Contained = %1.0f" % contained)
print("Punto tensione ....................")
print(pt)
print("Centro ....................")
print(center)



from pathlib import Path
import pytest

from pycivil.EXAStructural.loads import (
    ForcesOnSection,
)
from pycivil.EXAStructural.loads import Frequency_Enum as Fr
from pycivil.EXAStructural.loads import LimiteState_Enum as Ls
from pycivil.EXAUtils.latexReportMakers import ForcesOnSectionListFB
from pycivil.EXAUtils.report import (
    ReportDriverEnum,
    Reporter,
    ReportTemplateEnum,
    getTemplatesPath,
)

@pytest.mark.needLatex
def test_report_latex_loads(tmp_path: Path):

    latexTemplatePath = getTemplatesPath()
    fragmentBuilder = ForcesOnSectionListFB(latexTemplatePath)
    listOfForces = fragmentBuilder.forces()

    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 101,
            Fy=1e3 * 201,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f1",
            id=1,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 102,
            Fy=1e3 * 202,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f2",
            id=2,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 103,
            Fy=1e3 * 203,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f3",
            id=3,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 104,
            Fy=1e3 * 204,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f4",
            id=4,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 105,
            Fy=1e3 * 205,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f5",
            id=5,
            limitState=Ls.SERVICEABILITY,
            frequency=Fr.CHARACTERISTIC,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 106,
            Fy=1e3 * 206,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f6",
            id=6,
            limitState=Ls.SERVICEABILITY,
            frequency=Fr.FREQUENT,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 107,
            Fy=1e3 * 207,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f7",
            id=7,
            limitState=Ls.SERVICEABILITY,
            frequency=Fr.QUASI_PERMANENT,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 108,
            Fy=1e3 * 208,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f9",
            id=8,
            limitState=Ls.LIMIT_STATE_ND,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 101,
            Fy=1e3 * 201,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f1",
            id=1,
            frequency=Fr.FREQUENCY_ND,
            limitState=Ls.ULTIMATE,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 102,
            Fy=1e3 * 202,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f2",
            id=2,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 103,
            Fy=1e3 * 203,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f3",
            id=3,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 104,
            Fy=1e3 * 204,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f4",
            id=4,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 105,
            Fy=1e3 * 205,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f5",
            id=5,
            limitState=Ls.SERVICEABILITY,
            frequency=Fr.CHARACTERISTIC,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 106,
            Fy=1e3 * 206,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f6",
            id=6,
            limitState=Ls.SERVICEABILITY,
            frequency=Fr.FREQUENT,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 107,
            Fy=1e3 * 207,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f7",
            id=7,
            limitState=Ls.SERVICEABILITY,
            frequency=Fr.QUASI_PERMANENT,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 108,
            Fy=1e3 * 208,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f9",
            id=8,
            limitState=Ls.LIMIT_STATE_ND,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 101,
            Fy=1e3 * 201,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f1",
            id=1,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 102,
            Fy=1e3 * 202,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f2",
            id=2,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 103,
            Fy=1e3 * 203,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f3",
            id=3,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=1e3 * 104,
            Fy=1e3 * 204,
            Fz=1e3 * 300,
            Mx=1e3 * 100000,
            My=1e3 * 200000,
            Mz=1e3 * 300000,
            descr="f4",
            id=4,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )

    f1 = fragmentBuilder.buildFragment()

    print(f1.frags())

    reporter = Reporter(latexTemplatePath)
    reporter.linkFragments(template=ReportTemplateEnum.TEX_ENG_CAL, fragments=[f1])
    reporter.compileDocument(path=str(tmp_path))
    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.pdf").exists()
    assert tmp_path.joinpath("report.tex").exists()
